from pydtnn.backends.gpu.libs import libcudnn as cudnn

from pydtnn.backends.gpu.layers.abstract.block_layer import AbstractBlockLayerGPU
from pydtnn.layers.encoder import Encoder
from pydtnn.layers.decoder import Decoder
from pydtnn.layers.encoder_decoder import EncoderDecoder


class EncoderDecoderGPU(AbstractBlockLayerGPU, EncoderDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = [Encoder(embedl=self.embedl, d_k=self.d_k, d_ff=self.d_ff, heads=self.heads, dropout_rate=self.dropout_rate) for _ in range(self.enc_layers)]
        self.decoder = [Decoder(embedl=self.embedl, d_k=self.d_k, d_ff=self.d_ff, heads=self.heads, dropout_rate=self.dropout_rate) for _ in range(self.dec_layers)]
        self.paths = [self.encoder + self.decoder]

        # The next attributes will be initialized later
        self.y = self.dx = None

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        if len(x) == 2:
            x_enc, x_dec = x
            mask_enc = mask_dec = None
            enc_shape = (prev_shape[0], ())
            dec_shape = (prev_shape[0], prev_shape[1], ())
        else:
            x_enc, mask_enc, x_dec, mask_dec = x
            enc_shape = (prev_shape[0], prev_shape[1])
            dec_shape = (prev_shape[2], prev_shape[0], prev_shape[3])
        self.embedl = x_enc.shape[-1]

        # Initialize all sublayers
        for layer in self.children:
            layer.init_backend_from_model(self.model)

        self.encoder[0].initialize(prev_shape=enc_shape, x=(x_enc, mask_enc))
        for layer in self.encoder[1:]:
            layer.initialize(prev_shape=enc_shape, x=(x_enc, mask_enc))
        for layer in self.decoder:
            layer.initialize(prev_shape=dec_shape, x=(x_dec, x_enc, mask_dec))

        for layer in self.children:
            self.fwd_time += layer.fwd_time
            self.bwd_time += layer.bwd_time
            self.nparams += layer.nparams

    def initialize_block_layer(self):
        pass

    def forward(self, x):
        if len(x) == 2:
            x, y = x
            x_mask = y_mask = None
        else:
            x, x_mask, y, y_mask = x
        for i in range(self.enc_layers):  # Encoding layers
            x = self.encoder[i].forward(x, x_mask)
        for i in range(self.dec_layers):  # Decoding layers
            y = self.decoder[i].forward(y, x, y_mask)
        self.y = y
        return self.y

    def backward(self, prev_dx):
        alpha, beta = 1.0, 1.0
        dx_tgt, dx_enc = self.decoder[self.dec_layers-1].backward(prev_dx)
        for i in range(self.dec_layers-1, 0, -1):  # Decoding layers
            dx_tgt, dx_2 = self.decoder[i-1].backward(dx_tgt)
            cudnn.cudnnAddTensor(self.model.cudnn_handle,
                                 alpha, dx_2.desc, dx_2.ptr,
                                 beta, dx_enc.desc, dx_enc.ptr)  # dx_enc += dx2
        for i in range(self.enc_layers, 0, -1):  # Enconding layers
            dx_enc = self.encoder[i-1].backward(dx_enc)
        if self.need_dx:
            return dx_tgt, dx_enc



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def pointwise_conv_cython[T: _npDT](x: _npDT_4Dims[T], k: _npDT_2Dims[T], out: _npDT_4Dims[T]) -> None:
    """
    Args:
        x (npDT_4Dims): 4-dimensinal array where the input data is stored.
        k (npDT_2Dims): 2-dimensinal array where the kernel is stored.
        out (npDT_4Dims): 4-dimensinal array where the output is stored.
    Returns:
        Nothing. The output is stored in \"out\".
    """





from collections.abc import Callable, Sequence
from pydtnn.layer_base import LayerBase
from pydtnn.utils import find_component


def select(name: str) -> Callable[[Sequence[int], Sequence[int]], Sequence[LayerBase]]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



from collections.abc import Sequence
from pydtnn.layer_base import LayerBase
from pydtnn.layers.input import Input
from pydtnn.layers.multi_head_attention import MultiHeadAttention


def transformer_test(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=((1, 75, 300))))
    _(MultiHeadAttention(embedl=300, d_k=8, heads=4, dropout_rate=0.0))
    # _(LayerNormalization())
    # _(EncoderDecoder(enc_layers=6, dec_layers=6, embedl=300, d_k=30, heads=10, d_ff=1200, dropout_rate=0.0))

    return model



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def depthwise_conv_nhwc_cython[T: _npDT](x: _npDT_4Dims[T],
                                         k: _npDT_3Dims[T],
                                         res: _npDT_4Dims[T],
                                         ho: int, wo: int,
                                         vpadding: int, hpadding: int,
                                         vstride: int, hstride: int,
                                         vdilation: int, hdilation: int) -> None:
    """
    Args:
        x (npDT_4Dims): The 4 dimensional input's ndarray.
        k (npDT_3Dims): The 3dimensions ndarray that contains the kernel.
        res (npDT_4Dims): The 4 dimensional output's ndarray. Must be filled with zeros.
        ho: (int): Output's height value.
        wo: (int): Output's width value.
        vpadding (int): vertical padding value.
        hpadding (int): horizontal padding value.
        vstride (int): vertical stride value.
        hstride (int): horizontal stride value.
        vdilation (int): vertical dilation value.
        hdilation (int): horizontal dilation value.

    Returns:
        Nothing. The value is stores in \"res\".
    """


def depthwise_conv_backward_nhwc_cython[T: _npDT](dy: _npDT_4Dims[T],
                                                  x: _npDT_4Dims[T],
                                                  k: _npDT_3Dims[T],
                                                  dx: _npDT_4Dims[T],
                                                  dw: _npDT_3Dims[T],
                                                  vpadding: int, hpadding: int,
                                                  vstride: int, hstride: int,
                                                  vdilation: int, hdilation: int) -> None:
    """
    Args:
        dy (npDT_4Dims): The 4 dimensional array that contains the gradient of the backward's input.
        x (npDT_4Dims): The 4 dimensional array that contains the input forward's.
        k (npDT_3Dims): The 3 dimensional array that contains the kernel.
        dx npDT_4Dims: The 4 dimensional array that contains the input forward's gradient. Must be filled with zeros.
        dw npDT_3Dims: The 3 dimensional array that contains the kernel's gradient
        vpadding (int): vertical padding value.
        hpadding (int): horizontal padding value.
        vstride (int): vertical stride value.
        hstride (int): horizontal stride value.
        vdilation (int): vertical dilation value.
        hdilation (int): horizontal dilation value.

    Returns:
        Nothing. The outputs are stored in \"dx\" and \"dw\".
    """



import numpy as np
import pycuda.gpuarray as gpuarray  # type: ignore
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.elementwise import ElementwiseKernel  # type: ignore

from pydtnn.backends.gpu.optimizers.optimizer import OptimizerGPU
from pydtnn.optimizers.rmsprop import RMSProp
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE


class RMSPropGPU(RMSProp[TensorGPU], OptimizerGPU):
    """
    RMSPropGPU Optimizer
    """

    def __init__(self, learning_rate=1e-2, rho=0.9, epsilon=1e-7, decay=0.0, dtype=np.dtype(np.float32)):
        super().__init__(learning_rate, rho, epsilon, decay, dtype)

        pow_func = {np.dtype(np.float32): "powf", np.dtype(np.float64): "pow"}[dtype]

        # --- GPU ---
        parameters_gpu = "{T} *w, {T} *dw, {T} *cache, float lr, float decay, float rho, float epsilon".format(T=DTYPE2CTYPE[dtype])
        operations_gpu = "cache[i] = rho * cache[i] + (1 - rho) * {func}(dw[i], 2); \
                                             w[i] -= lr * (decay * w[i] + (dw[i] / sqrtf(cache[i] + epsilon)))".format(func=pow_func)
        self.update_gpu = ElementwiseKernel(parameters_gpu, operations_gpu, "RMSProp_kernel")
        # -----------

        # GPU DIRECT -
        _name = "RMSProp_kernel_gpudirect"
        code = """
        __global__ void {name}({T} *w, {T} *dw, {T} *cache,
                                float lr, float decay, float rho, float epsilon, int N) 
        {{
                int i = blockIdx.x * blockDim.x + threadIdx.x;
                if (i < N) {{
                    cache[i] = rho * cache[i] + (1 - rho) * {func}(dw[i], 2);
                    w[i] -= lr * (decay * w[i] + (dw[i] / sqrt(cache[i] + epsilon)));
                }}
        }}
        """.format(T=DTYPE2CTYPE[dtype],
                   func=pow_func,
                   name = _name
                   )
        self.update_gpudirect = SourceModule(code).get_function(_name)
        # -------------

    def initialize(self, list_layers: list[LayerGPU]) -> None:
        for layer in list_layers:
            list_grad_vars = list(layer.grad_vars.keys())

            if len(list_grad_vars) != 0:
                self.context[layer.id] = dict[str, gpuarray.GPUArray]()
                for w_ in list_grad_vars:
                    w = getattr(layer, w_)
                    self.context[layer.id]["cache_%s" % w_] = gpuarray.zeros_like(w.ary, dtype=layer.model.dtype)

    def update(self, layer: LayerGPU):
        for w_, dw_ in layer.grad_vars.items():
            w, dw = getattr(layer, w_), getattr(layer, dw_)
            cache = self.context[layer.id]["cache_%s" % w_]
            w: TensorGPU
            dw: TensorGPU
            cache: gpuarray.GPUArray

            if self.gpudirect:
                n = self.get_batch_size(w)
                threads, blocks = self.get_threads_and_blocks()

                self.update_gpudirect(w.ary.gpudata, dw.ptr_intp, cache.gpudata,
                                      np.float32(self.learning_rate),
                                      np.float32(self.decay), np.float32(self.rho),
                                      np.float32(self.epsilon), np.int32(n),
                                      grid=(int(blocks), 1, 1), block=(int(threads), 1, 1),
                                      stream=layer.stream_2)
            else:
                self.update_gpu(w.ary, dw.ary, cache, np.float32(self.learning_rate),
                                np.float32(self.decay), np.float32(self.rho),
                                np.float32(self.epsilon), stream=layer.stream_2)



import time
from collections import defaultdict

import numpy as np

from pydtnn.tracers.simple_tracer import SimpleTracer
from pydtnn.utils.pmlib import PMLib

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pympi.MPI import Comm as MPI_COMM
else:
    from types import ModuleType
    MPI_COMM = ModuleType


class SimpleTracerPMLib(SimpleTracer):
    """
    SimpleTracerPMLib
    """

    def __init__(self, tracing: bool, output_filename: str, comm: MPI_COMM | None, pmlib_server_ip: str, pmlib_port: int, pmlib_device: str):
        super().__init__(tracing, output_filename, comm)
        if self.rank == 0:
            self.pmlib = PMLib(pmlib_server_ip, pmlib_port, verbose=True)
            self.pmlib_device = pmlib_device
        self.times = defaultdict(lambda: defaultdict(lambda: []))
        self.pending_times = []

    def enable_tracing(self):
        super().enable_tracing()
        # Start counter
        if self.rank == 0:
            self.pmlib.create_counter(self.pmlib_device)
            self.pmlib.start_counter()

    def _emit_event(self, evt_type_val: int, evt_val: int, stream=None):
        """This method will be called only if tracing is enabled"""
        super()._emit_event(evt_type_val, evt_val, stream)
        if evt_val != 0:
            self.pending_times.append((evt_type_val, evt_val, time.time()))
        else:
            end_time = time.time()
            _evt_type_val, _evt_val, start_time = self.pending_times.pop()
            self.times[_evt_type_val][_evt_val].append((start_time, end_time))

    def _output_header(self) -> str:
        output = super()._output_header()
        output += ";Joules"
        for i in range(1, self.pmlib.len_lines):
            output += f";Line{i - 1}"
        return output + ";Mean of intermediate power samples"

    def _output_row(self, event_type_value: int, event_value: int) -> str:
        output = super()._output_row(event_type_value, event_value) + ";"
        joules = np.zeros(self.pmlib.len_lines)
        intermediate_samples = 0
        for start_time, end_time in self.times[event_type_value][event_value]:
            joules += self.pmlib.get_joules(start_time, end_time)
            intermediate_samples += self.pmlib.get_number_of_intermediate_samples(start_time, end_time)
        if len(self.times[event_type_value][event_value]) > 0:
            intermediate_samples = intermediate_samples / len(self.times[event_type_value][event_value])
        output += ";".join([f"{x}" for x in joules])
        return output + f";{intermediate_samples}"

    def _write_output(self):
        """This method will be called at exit only if tracing has been enabled at any time"""
        if self.rank == 0:
            self.pmlib.stop_counter()
            self.pmlib.get_counter_data()
            super()._write_output()
            watts_filename = self.output_filename + ".watts"
            print(f"Writing watts output to '{watts_filename}'...")
            with open(watts_filename, 'w') as f:
                header = "Time"
                header += ";Watts"
                for i in range(1, self.pmlib.len_lines):
                    header += f";Line{i - 1}"
                f.write(header + "\n")
                elapsed_times = self.pmlib.times - self.pmlib.times[0]
                elapsed_times_watts = np.concatenate((elapsed_times.reshape(1, -1), self.pmlib.watts)).transpose()
                for row in elapsed_times_watts:
                    f.write(";".join(f"{x}" for x in row) + "\n")
            self.pmlib.finalize_counter()



import numpy as np

from pydtnn.layers.layer import Layer, LayerError
from pydtnn.utils.constants import Array


class AbstractPool2DLayer[T: Array](Layer[T]):

    def __init__(self, pool_shape: tuple[int, int] | int = (2, 2), padding: tuple[int, int] | int = 0,
                 stride: tuple[int, int] | int = 1, dilation: tuple[int, int] | int = 1):
        super().__init__()
        self.pool_shape = (pool_shape, pool_shape) if isinstance(pool_shape, int) else pool_shape
        self.padding = padding
        self.stride = stride
        self.dilation = dilation
        self.vpadding, self.hpadding = (padding, padding) if isinstance(padding, int) else padding
        self.vstride, self.hstride = (stride, stride) if isinstance(stride, int) else stride
        self.vdilation, self.hdilation = (dilation, dilation) if isinstance(dilation, int) else dilation
        self.ci = self.hi = self.wi = self.kh = self.kw = self.ho = self.wo = self.co = self.n = 0

    def initialize(self, prev_shape, x: T | None):
        super().initialize(prev_shape, x)
        self.ci, self.hi, self.wi = self.model.decode_shape(prev_shape)
        if self.pool_shape[0] == 0:
            self.pool_shape = (self.hi, self.pool_shape[1])
        if self.pool_shape[1] == 0:
            self.pool_shape = (self.pool_shape[0], self.wi)
        self.kh, self.kw = self.pool_shape
        self.co = self.ci
        self.ho = (self.hi + 2 * self.vpadding - self.vdilation * (self.kh - 1) - 1) // self.vstride + 1
        self.wo = (self.wi + 2 * self.hpadding - self.hdilation * (self.kw - 1) - 1) // self.hstride + 1
        if not (self.ho > 0 and self.wo > 0):
            raise LayerError(f"Output dimensions must be greater than 0. ho: {self.ho}, wo: {self.wo}.")
        self.shape = self.model.encode_shape((self.co, self.ho, self.wo))
        self.n = np.prod(self.shape)

    def _show_props(self) -> dict:
        props = super()._show_props()

        props["pool"] = self.pool_shape
        props["padding"] = (self.vpadding, self.hpadding)
        props["stride"] = (self.vstride, self.hstride)
        props["dilation"] = (self.vdilation, self.hdilation)

        return props



from numpy import int32, prod
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.optimizers.optimizer import Optimizer


class OptimizerGPU(Optimizer[TensorGPU]):
    """
    Extends an Optimizer class with the attributes and methods required by GPU Optimizers.
    """

    LIMIT_THREADS_AND_BLOCKS = 1024

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gpudirect = False

    def set_gpudirect(self, gpudirect: bool):
        self.gpudirect = gpudirect

    def get_batch_size(self, w: TensorGPU) -> int32:
        return int32(prod((self.model.real_batch_size, *(w.shape[1:]))))

    def get_threads_and_blocks(self):
        threads = min(self.model.real_batch_size, self.LIMIT_THREADS_AND_BLOCKS)
        blocks = max(self.model.real_batch_size, self.LIMIT_THREADS_AND_BLOCKS) // threads + 1
        return threads, blocks



import numpy as np

from pydtnn.layers.layer import Layer
from pydtnn.utils.constants import Array


# https://melfm.github.io/posts/2018-08-Understanding-Normalization/

class LayerNormalization[T: Array](Layer[T]):
    def __init__(self, axis=(-2, -1), beta: float = 0.0, gamma: float = 1.0,
                 epsilon: float = 1e-5,
                 sync_stats: bool = False):
        super().__init__()
        if type(axis) is not tuple:
            self.axis = (axis,)
        else:
            self.axis = axis
        self.gamma_init_val = gamma
        self.beta_init_val = beta
        self.epsilon = epsilon
        self.grad_vars = {"beta": "dbeta", "gamma": "dgamma"}
        self.sync_stats = sync_stats
        # The next attributes will be initialized later
        self.gamma = self.beta = None
        self.std = self.xn = None
        self.dgamma = self.dbeta = None

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        self.shape = shape_ = prev_shape
        self.gamma = np.full(shape_, self.gamma_init_val, self.model.dtype)
        self.beta = np.full(shape_, self.beta_init_val, self.model.dtype)
        self.nparams = self.gamma.size + self.beta.size



from pathlib import Path

from setuptools import setup, Extension

import numpy
from Cython.Build import cythonize


setup(
    ext_modules=cythonize([
        Extension(
            ".".join(pyx.with_suffix("").parts),
            [str(pyx)],
            extra_compile_args=["-fopenmp", "-O3", "-march=native", "-g0"],
            extra_link_args=["-fopenmp"],
            include_dirs=[numpy.get_include()],
        ) for pyx in Path("pydtnn").rglob("*.pyx")
    ], language_level=3),
)



"""
PyDTNN model
"""

import enum
import importlib
import itertools
from math import ceil
import operator
import time
from functools import cached_property, reduce
from timeit import default_timer as timer
from types import ModuleType
from typing import TYPE_CHECKING, Any, Literal
from warnings import warn
from collections import abc

import numpy as np
from tqdm import tqdm

# TODO: Check if all the elements imported here are necessary and if they are corretly set in Model's code.
from pydtnn import MPI_MODULE, Cudnn_Handle_Type, Cublas_Handle_Type, gpu_errors, MPI, drv, gpuarray, tensor_gpu, nccl, cudnn, cublas, rank, nprocs, hostname, ranks_per_node, num_gpus, supported_gpu, nccl_comm, cudnn_handle, cublas_handle, device, context, stream

from pydtnn import utils
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.activations.relu import Relu
from pydtnn.backends import BackendType
from pydtnn.backends.gpu.optimizers.optimizer import OptimizerGPU
from pydtnn.libs.libmpi import proto as PROTOCOL
from pydtnn.libs import libcrypt
from pydtnn.datasets.dataset import Dataset
from pydtnn.layer_base import LayerBase, FusedLayerMixIn
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.losses.loss import Loss
from pydtnn.datasets.dataset import select as select_dataset
from pydtnn.losses.loss import select as select_loss
from pydtnn.optimizers.optimizer import select as select_optimizer
from pydtnn.metrics.metric import select as select_metric
from pydtnn.models.model import select as select_model
from pydtnn.schedulers.scheduler import select as select_scheduler
from pydtnn.layers.layer import select as select_layer
from pydtnn.parser import PydtnnArgumentParser
from pydtnn.utils.performance_models import allreduce_time
from pydtnn.tracers.events import PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum
from pydtnn.tracers.extrae_tracer import ExtraeTracer
from pydtnn.tracers.simple_tracer import SimpleTracer
from pydtnn.tracers.simple_tracer_gpu import SimpleTracerGPU
from pydtnn.tracers.simple_tracer_pmlib import SimpleTracerPMLib
from pydtnn.tracers.tracer import Tracer
from pydtnn.utils.best_of import BestOf
from pydtnn.utils.memory_cache import MemoryCache
from pydtnn.utils.performance_counter import PerformanceCounter
from pydtnn.utils.tensor import SampleFormat, TensorFormat, format_reshape, encode_shape, encode_tensor, decode_shape, decode_tensor
from pydtnn.utils.constants import Array, NetworkAlgEnum, ArrayShape, Parameters
from pydtnn.metrics.metric import Metric


# --- CONSTANS --- #
BAR_WIDTH = 140
DEFAULT_BACH_SIZE = 64

# NOTE: mpi4py has more functions, but no typing
if TYPE_CHECKING:
    from pympi.MPI import Comm as MPI_COMM
else:
    MPI_COMM = ModuleType


def get_tracer(tracer_output: str, tracing: bool, comm: MPI_COMM | None, enable_gpu: bool,
               tracer_pmlib_server: str, tracer_pmlib_port: int, tracer_pmlib_device: str) -> Tracer:

    if tracer_output == "":
        tracer = ExtraeTracer(tracing)
    else:
        if enable_gpu:
            tracer = SimpleTracerGPU(tracing, tracer_output, comm)
        else:
            if tracer_pmlib_device != "":
                tracer = SimpleTracerPMLib(tracing, tracer_output, comm, tracer_pmlib_server, tracer_pmlib_port, tracer_pmlib_device)
            else:
                tracer = SimpleTracer(tracing, tracer_output, comm)
    return tracer


def get_tensor_format(tensor_format: Literal["AUTO", "NCHW", "NHWC"] = "AUTO", gpu: bool = False) -> TensorFormat:
    match tensor_format.upper():
        case "AUTO":
            return TensorFormat.NCHW if gpu else TensorFormat.NHWC
        case "NCHW":
            return TensorFormat.NCHW
        case "NHWC":
            return TensorFormat.NHWC
        case _:
            raise NotImplementedError(f"\'{tensor_format}\' is not supported.")


def get_batch_size(local_size: int | None, global_size: int | None, comm_size: int, default: int = DEFAULT_BACH_SIZE) -> int:
    if local_size and global_size:
        raise ValueError("Can not define 'local_batch_size' and 'global_batch_size' simultaneously")

    if global_size:
        # NOTE: Using comm_size instead of nprocs might not be appropriate,
        #       as it differs to how global_batch_size is defined elsewhere,
        #       but for now it just a parser option difference that helps testing
        batch_size = global_size // comm_size
    elif local_size:
        batch_size = local_size
    else:
        batch_size = default

    if batch_size < 1:
        raise ValueError(f"'batch_size' ({batch_size}) too small or too many processes (num processes: {comm_size})")

    return batch_size


class CudnnDataType(enum.StrEnum):
    FLAOT64 = "CUDNN_DATA_DOUBLE"
    FLOAT32 = "CUDNN_DATA_FLOAT"
    INT8 = "CUDNN_DATA_INT8"
    INT32 = "CUDNN_DATA_INT32"


class Model[T: Array]:
    """
    PyDTNN Model
    """

    class ParallelMode(enum.StrEnum):
        SEQUENTIAL = enum.auto()
        DATA = enum.auto()

    class Mode(enum.StrEnum):
        EVALUATE = enum.auto()
        TRAIN = enum.auto()
        UNSPECIFIED = enum.auto()

    class SyncAlg(enum.StrEnum):
        AVG = enum.auto()
        WAVG = enum.auto()
        INVAVG = enum.auto()

    class SyncParticipation(enum.StrEnum):
        ALL = enum.auto()
        AVAIL2ALL = enum.auto()

# Explicit declaration of those model attributes that are referenced by other parts of PyDTNN
#   NOTE: The following parameters come from "Parser"
    steps_per_epoch: int
    cpu_speed: float
    memory_bw: float
    network_bw: float
    network_lat: float
    network_alg: NetworkAlgEnum
    loss_func_name: str
    num_epochs: int
    model_sync_freq: int
    final_model_sync: bool
    test_as_validation: bool
    validation_split: float
    use_synthetic_data: bool
    dataset_train_path: str
    dataset_test_path: str
    evaluate_only: bool
    evaluate_on_train: bool
    profile: bool
    history_file: str
    model_sync_min_avail: int
    dataset_name: str
    shared_storage: bool
    encryption_name: str
    augment_flip: bool
    augment_flip_prob: float
    augment_crop: bool
    augment_crop_size: int
    augment_crop_prob: float
    transform_crop: bool
    transform_crop_perc: int
    transform_resize: bool
    transform_resize_dsize: int
    initial_model_sync: bool
    dataset_percentage: float
    use_mpi_buffers: bool
    enable_memory_cache: bool
    gpus_per_node: int
    weights_and_bias_filename: str
    learning_rate_scaling: bool
    metrics: str
# ------------

    rank_weight: float
    comm_rank: int
    comm_size: int
    rank: int
    nprocs: int
    learning_rate: float
    MPI: MPI_MODULE | None
    comm: MPI_COMM | None

    nccl_type: Any | None
    nccl_comm: Any | None
    cudnn_handle: Cudnn_Handle_Type | None
    cublas_handle: Cublas_Handle_Type | None
    stream: Any  # drv.Stream
    cudnn_dtype: int
    input_shape: ArrayShape
    output_shape: ArrayShape

    dtype: np.dtype

    real_batch_size: int
    nparams: int

    y_batch: T
    history: dict[str, list[np.ndarray]]
    loss_func: Loss
    metrics_funcs: list[Metric]
    loss_and_metrics: list[str]
    total_metrics: np.ndarray

    cuda_grid: tuple[int, int, int]
    cuda_block: tuple[int, int, int]

    def __init__(self, parallel: ParallelMode = ParallelMode.SEQUENTIAL, use_blocking_mpi: bool = False, enable_gpu: bool = False,
                 enable_gpudirect: bool = False, enable_nccl: bool = False, dtype: np.dtype = np.dtype(np.float32), tracing: bool = False,
                 tracer_output: str = "", tracer_pmlib_server: str = "127.0.0.1", tracer_pmlib_port: int = 6526,
                 tracer_pmlib_device: str = "", **kwargs):

        # Attributes related to the given arguments
        self.parallel: Model.ParallelMode = Model.ParallelMode(parallel)
        self.blocking_mpi: bool = use_blocking_mpi
        self.enable_gpu = self.enable_cudnn = enable_gpu
        self.gpudirect: bool = enable_gpudirect
        self.enable_nccl: bool = enable_nccl
        self.dtype: np.dtype = np.dtype(dtype)

        self._sync_x_y = self._sync_x_y_gpu if self.enable_gpu else self._sync_x_y_cpu  # type: ignore

        self.nparams = 0

        # Get default values from parser and update them from the received kwargs
        self.kwargs: dict[str, Any] = PydtnnArgumentParser().get_default_values()
        self.kwargs.update(kwargs)

        # NOTE: self.conv_variant comes from Parser
        self.conv_variant = Conv2D.Variant[self.conv_variant.upper()]

        # Set MPI and comm
        self._init_comms()

        # Set tracer
        self.tracer = get_tracer(tracer_output=tracer_output, tracing=tracing, comm=self.comm, enable_gpu=enable_gpu,
                                 tracer_pmlib_server=tracer_pmlib_server, tracer_pmlib_port=tracer_pmlib_port,
                                 tracer_pmlib_device=tracer_pmlib_device)

        # Set performance counter
        self.perf_counter = PerformanceCounter()

        # Layers' attributes
        self.layers: list[LayerBase] = []
        self.layer_id_generator: abc.Iterator[int] = iter(itertools.count())

        # Matmul
        self.matmul = utils.matmul

        # Set current mode to unspecified
        self.mode: Model.Mode = Model.Mode.UNSPECIFIED

        # Memory cache optimization
        if self.enable_memory_cache:
            MemoryCache.enable()
        else:
            MemoryCache.disable()

        # Cuda
        if self.enable_cudnn:
            if gpuarray and drv and cublas:
                self._initialize_cuda()
            else:
                raise ExceptionGroup("CUDA import error", gpu_errors)

        # Data format
        self.tensor_format: TensorFormat = get_tensor_format(tensor_format=self.tensor_format, gpu=self.enable_cudnn)  # type: ignore

        # Disable BestOf globally if not enabled
        if self.enable_best_of is False:
            BestOf.use_always_the_first_alternative()

        self.batch_size = get_batch_size(local_size=self.batch_size, global_size=self.global_batch_size, comm_size=self.comm_size)

        # Attributes that will be properly initialized elsewhere

        # ---

        # Encryption
        if self.encryption_name:
            self.crypt = self._init_crypt(self.encryption_name)

        else:
            self.crypt = None

        # Load weights and bias
        if self.weights_and_bias_filename:
            self.load_weights_and_bias(self.weights_and_bias_filename)
        # Dataset
        if self.dataset_name:
            self.dataset: Dataset = select_dataset(self.dataset_name)(self)

        # Optimizers and LRSchedulers
        if self.learning_rate_scaling:
            # using comm_size instead of nprocs might not be appropriate,
            # as it differs to how learning_rate is defined elsewhere,
            # but for now it just a parser option difference that helps testing
            self.learning_rate = self.learning_rate / self.comm_size

        self.optimizer = select_optimizer(self.optimizer_name).from_model(self)
        self.optimizer.init_backend_from_model(self)

        self.schedulers = [
            select_scheduler(scheduler_name).from_model(self)
            for scheduler_name in filter(None, self.schedulers_names.split(","))
        ]
        for scheduler in self.schedulers:
            scheduler.set_model(self)

        # Metrics list
        self.metrics_list: list[str] = [m for m in self.metrics.replace(" ", "").split(",")]

        # Private attributes
        self._evaluate_round: int = 0
        self._initialized: bool = False

        # Read the model (must be the last action, as it calls self._initialize() if there is a model)
        self.model_name: str | None = self.kwargs.get("model_name")
        if self.model_name:
            self._read_model(self.model_name)

        # Syncronization parameters
        # NOTE: This parameter come from Parser.
        self.model_sync_alg = Model.SyncAlg(self.model_sync_alg)

        # NOTE: This parameter come from Parser.
        self.model_sync_participation = Model.SyncParticipation(self.kwargs["model_sync_participation"])

    def _init_comms(self) -> None:
        # Comunication type
        match self.parallel:
            case "sequential":
                self.MPI, self.comm = (None, None)
            case "data":
                if not MPI:
                    raise ValueError("Please, install mpi4py to allow parallel MPI execution!")
                self.MPI, self.comm = (MPI, MPI.COMM_WORLD)
            case _:
                raise ValueError(f"Parallel option '{self.parallel}' not recognized.")

        # Comunication size
        self.rank_weight = 1.0
        self.comm_rank = self.rank = 0
        self.comm_size = self.nprocs = 1
        if self.comm:
            self.comm_rank = self.comm.Get_rank()
            self.comm_size = self.comm.Get_size()
            if self.shared_storage:
                self.rank = self.comm_rank
                self.nprocs = self.comm_size

        # Comunication method
        match self.use_mpi_buffers:
            case None:
                self.use_mpi_buffers = PROTOCOL is None
            case bool():
                pass
            case _:
                raise ValueError(f"MPI buffers option '{self.use_mpi_buffers}' not recognized.")

    def _initialize_cuda(self) -> None:
        LIMIT_THREADS_AND_BLOCKS = 1024
        self.cuda_threads = min(self.batch_size, LIMIT_THREADS_AND_BLOCKS)
        self.cuda_blocks = (max(self.batch_size, LIMIT_THREADS_AND_BLOCKS) // self.cuda_threads) + 1
        # NOTE: Seems that in PyDTNN, usually the ".x" (blockIdx.x, threadIdx.x, ...) is the only dimension used.
        self.cuda_grid = (self.cuda_blocks, 1, 1)
        self.cuda_block = (self.cuda_threads, 1, 1)

        assert drv is not None
        assert context is not None
        assert cudnn_handle is not None
        assert cublas_handle is not None
        assert stream is not None

        if not self.gpudirect and self.enable_nccl:
            raise RuntimeError("It is necessary to have gpudirect active to work with NCCL.")

        if self.comm and self.enable_nccl:
            assert nccl is not None
            assert nccl_comm is not None

            nccl_types = {np.float64: nccl.DataType.Float64,
                          np.float32: nccl.DataType.Float32,
                          np.int8: nccl.DataType.Int8,
                          np.int32: nccl.DataType.Int32}

            nccl_type = nccl_types.get(self.dtype, nccl.DataType.Float32)

            if ranks_per_node[hostname] > num_gpus:
                raise ValueError("Not able to run more processes than GPUs per node!")
        else:
            nccl_type = None

        self.tracer.set_stream(stream)

        cudnn_types = {np.float64: CudnnDataType.FLAOT64,
                       np.float32: CudnnDataType.FLOAT32,
                       np.int8: CudnnDataType.INT8,
                       np.int32: CudnnDataType.INT32}

        cudnn_type: str = cudnn_types.get(self.dtype, CudnnDataType.FLOAT32)
        cudnn_dtype: int = cudnn.cudnnDataType[cudnn_type]

        self.nccl_type = nccl_type
        self.nccl_comm = nccl_comm
        self.cudnn_handle = cudnn_handle
        self.cublas_handle = cublas_handle
        self.stream = stream
        self.cudnn_dtype = cudnn_dtype

    def _ensure_model_runnable(self) -> None:
        if not self._initialized:
            self._initialize()
        are_layers = bool(self.layers)
        if not are_layers:
            warn("The model has no layers in it.", RuntimeWarning)
        elif not self.dataset:
            raise ValueError("There is no dataset and the model has layers.")

    @cached_property
    def empty_x(self) -> TensorGPU:
        # NOTE: Can not allocate a zero-size array, so slice one
        assert gpuarray and self.cudnn_dtype
        empty_x = gpuarray.empty((1, *self.dataset.input_shape), self.dtype)[:0]
        return TensorGPU(empty_x, self.tensor_format, self.cudnn_dtype)

    @cached_property
    def empty_y_tag(self) -> TensorGPU:
        # NOTE: Can not allocate a zero-size array, so slice one
        assert gpuarray and self.cudnn_dtype
        empty_y_tag = gpuarray.empty((1, *self.dataset.output_shape), self.dtype)[:0]
        return TensorGPU(empty_y_tag, self.tensor_format, self.cudnn_dtype)

    @property
    def dataset_path(self) -> str:
        """Raw dataset path with rank substituted"""
        return utils.string_substitute(self.kwargs["dataset_path"], rank=self.comm_rank)

    def __getattr__(self, item) -> Any:
        return self.kwargs.get(item)

    def _init_crypt(self, encryption_name: str) -> libcrypt.Context:
        """Inizialize encryption context"""
        try:
            module = importlib.import_module(f"pydtnn.libs.libcrypt.{encryption_name}")
        except Exception as exc:
            raise ValueError(f"Unsupported encryption module {encryption_name}!") from exc

        if self.comm_rank == 0:
            crypt = module.Context(
                poly_degree=self.encryption_poly_degree,
                global_scale=self.encryption_global_scale,
                security_level=self.encryption_security_level
            )

        if self.comm:
            crypt = self.comm.bcast(crypt if self.comm_rank == 0 else None)

        assert crypt is not None
        if self.enable_nccl:
            warn("If NCCL is active, encryption is disabled", RuntimeWarning)

        return crypt

    def _read_model(self, model_name: str) -> None:
        create_model = select_model(model_name)

        # NOTE: Dataset is always in NCHW
        # Change input_shape to model.tensor_format
        input_shape = format_reshape(self.dataset.input_shape, SampleFormat.CHW, self.tensor_format.as_sample())  # type: ignore
        if len(input_shape) != 3:
            warn(f"Input layer does not have 3 dimensions ({input_shape}), it may cause issues!", RuntimeWarning)
        launch_shape_warning = len(input_shape) == 3 and not (input_shape[0] > input_shape[2]) if self.tensor_format is TensorFormat.NHWC \
            else len(input_shape) == 3 and not (input_shape[0] < input_shape[1])
        if launch_shape_warning:
            warning_text = f"Input layer shape {input_shape} may not be in {self.tensor_format} format, regardless of model format! "
            warn(warning_text, RuntimeWarning)
            warning_text = None
        output_shape = tuple(self.dataset.output_shape)

        self.input_shape = input_shape
        self.output_shape = output_shape

        layers = create_model(input_shape, output_shape)
        self.add_layers(layers)  # type: ignore

        self._initialize()

    def encode_shape(self, shape: ArrayShape) -> ArrayShape:
        """Transform the shape from `NCHW` order to `model.tensor_format` order (supports 4 or 3 dimensions)"""
        return encode_shape(shape, self.tensor_format)

    def decode_shape(self, shape: ArrayShape) -> ArrayShape:
        """Transform the shape from `model.tensor_format` order to `NCHW` order (supports 4 or 3 dimensions)."""
        return decode_shape(shape, self.tensor_format)

    def encode_tensor(self, data: T) -> T:
        """Transpose elements of data from `NCHW` format to `model.tensor_format` format (supports 4 or 3 dimensions)."""
        return encode_tensor(data, self.tensor_format)  # type: ignore (TensorGPU does not have transpose yet)

    def decode_tensor(self, data: T) -> T:
        """Transpose elements of data from `model.tensor_format` format to `NCHW` format (supports 4 or 3 dimensions)."""
        return decode_tensor(data, self.tensor_format)  # type: ignore (TensorGPU does not have transpose yet)

    def _show_props(self) -> dict:
        props = {}

        if self.model_name:
            props["name"] = self.model_name

        if self.dataset_name:
            props["dataset"] = self.dataset_name

        if self.nparams > 0:
            props["params"] = self.nparams
            props["memory"] = utils.convert_size_bytes(self.nparams * self.dtype.itemsize)

        if self.layers:
            props["input"] = self.layers[0].shape
            props["output"] = self.layers[-1].shape
            props["batch-size"] = self.batch_size
            props["layers"] = len(self.get_all_layers())

        return props

    def __repr__(self) -> str:
        props = " ".join(
            f"{key}={value!r}"
            for key, value in self._show_props().items()
        )

        return f"<{self.__class__.__name__} {props}>"

    def show_layers(self) -> None:
        struct: dict[str, int] = {}
        all_props = {
            layer.id: layer._show_props()
            for layer in self.get_all_layers()
        }

        # Calculate headers and sizes
        for props in sorted(all_props.values(), key=lambda props: (-len(props), *props)):
            for key, value in props.items():
                struct[key] = max(struct.get(key, len(key)), len(str(value)))

        # Add header padding
        for header, size in struct.items():
            struct[header] += 2

        # Generate separator
        sep = ""
        for header, size in struct.items():
            sep += "+" + "-" * size
        sep += "+"

        # Show header
        print(sep)
        for header, size in struct.items():
            print(f"|{header.title():^{size}s}", end="")
        print("|")

        # Show layers
        top_layers = {layer.id for layer in self.layers}
        for layer_id, props in all_props.items():
            if layer_id in top_layers:
                print(sep)
            for header, size in struct.items():
                value = props.get(header, "")
                print(f"|{str(value):^{size}s}", end="")
            print("|")
        print(sep)

    def show_model(self) -> None:
        key = "Model Summary"
        print(key + "\n" + "=" * len(key))
        for key, value in self._show_props().items():
            print(f"- {key.replace('-', ' ').capitalize()}: {value}")

    def show(self) -> None:
        self.show_model()
        print()
        self.show_layers()
        print()

    def print_in_convdirect_format(self) -> None:
        line = "#l\tkn\two\tho\tt\tkh\tkw\tci\twi\thi"
        print(line)
        for layer in self.layers:
            layer.print_in_convdirect_format()

    def add(self, layer: LayerBase[T]) -> None:
        layer.init_backend_from_model(self)

        if self.layers:
            prev_shape = self.layers[-1].shape
            y = self.layers[-1].y
        else:
            prev_shape = ()
            y = None

        layer.initialize(prev_shape, y)

        self.nparams += layer.nparams
        self.layers.append(layer)

        if layer.act:
            self.add(layer.act())

    def add_layers(self, list_layers: list[LayerBase[T]]) -> None:
        for layer in list_layers:
            self.add(layer)
    # --- END add_layers ---

    def get_all_layers(self, from_layers: list[LayerBase[T]] | None = None) -> list[LayerBase[T]]:
        if from_layers is None:
            from_layers = self.layers
        this_recursion_layers = []
        for layer in from_layers:
            this_recursion_layers.append(layer)
            children = layer.children
            this_recursion_layers += self.get_all_layers(children)
        return this_recursion_layers

    def _select_fusion_3(self, fused_layers: list) -> tuple[str, list[LayerBase]]:
        layer2 = fused_layers[-1] if len(fused_layers) > 0 else None
        layer1 = fused_layers[-2] if len(fused_layers) > 1 else None
        layer0 = fused_layers[-3] if len(fused_layers) > 2 else None

        layer_name = None

        match (layer0, layer1, layer2):
            case (_, FusedLayerMixIn(), _): pass  # else: layer_name = None
            case (Conv2D(), BatchNormalization(), Relu()):
                if self.enable_fused_conv_bn_relu:
                    layer_name = "conv_2d_batch_normalization_relu"
                # else: layer_name = None
            case default: pass  # else: layer_name = None

        return layer_name, [layer0, layer1, layer2]
    # ----

    def _select_fusion_2(self, fused_layers: list) -> tuple[str, list[LayerBase]]:
        layer2 = fused_layers[-1] if len(fused_layers) > 0 else None
        layer1 = fused_layers[-2] if len(fused_layers) > 1 else None

        layer_name = None

        match (layer1, layer2):
            case (FusedLayerMixIn(), _): pass  # else: layer_name = None
            # else: layer_name = None
            case (Conv2D(), BatchNormalization()):
                if self.enable_fused_conv_bn:
                    layer_name = "conv_2d_batch_normalization"
                # else: layer_name = None
            case (Conv2D(), Relu()):
                if self.enable_fused_conv_relu:
                    layer_name = "conv_2d_relu"
                # else: layer_name = None
            case (BatchNormalization(), Relu()):
                if self.enable_fused_bn_relu:
                    layer_name = "batch_normalization_relu"
                # else: layer_name = None
            case default: pass  # else: layer_name = None

        return layer_name, [layer1, layer2]
    # ----

    def __layer_fusion(self, layers: list[LayerBase], switch_fusion: abc.Callable) -> None:
        i = 0
        while i < len(layers):
            curr_layer = layers[i]

            # Recurse if layer group
            for j, p in enumerate(curr_layer.paths):
                self.__layer_fusion(curr_layer.paths[j], switch_fusion)

            layer_name, layers_to_fuse = switch_fusion(layers[:i])

            if layer_name:
                dict_params = reduce(operator.or_, (layer.__dict__ for layer in reversed(layers_to_fuse)))
                print(f"Fusing {' + '.join(map(lambda layer: layer.name_with_id, layers_to_fuse))}")
                fused_layer = select_layer(layer_name)

                new_curr_layer = fused_layer(from_parent=dict_params)  # type: ignore (it's okay)
                new_curr_layer.init_backend_from_model(self)
                new_curr_layer.__dict__.update(dict_params)
                try:
                    new_curr_layer.initialize(prev_shape=layers_to_fuse[0].prev_shape, x=layers_to_fuse[0].x)
                except Exception as e:
                    warn(f"Aborted fusion, {e}")
                else:
                    start = i - len(layers_to_fuse)
                    del layers[start: i]
                    layers.insert(start, new_curr_layer)
                    i -= len(layers_to_fuse)
            i += 1

    def _apply_layer_fusion(self):
        """ Apply layer fusion in a recursive manner """

        if not self.enable_cudnn and any([self.enable_fused_bn_relu, self.enable_fused_conv_relu, self.enable_fused_conv_bn, self.enable_fused_conv_bn_relu]):
            # NOTE: 1st the 3 layers fusion, then the rest:
            self.__layer_fusion(self.layers, self._select_fusion_3)
            self.__layer_fusion(self.layers, self._select_fusion_2)

    @property
    def _backend(self) -> BackendType:
        return BackendType.GPU if self.enable_gpu else BackendType.CPU

    def _initialize(self):
        if self._initialized:
            return
        self._apply_layer_fusion()
        loss_cls = select_loss(self.loss_func_name)
        self.loss_func = loss_cls(shape=(self.batch_size, *self.layers[-1].shape))
        self.loss_func.init_backend_from_model(self)
        self.loss_func.initialize()
        self.metrics_funcs = [select_metric(m)(shape=(self.batch_size, *self.layers[-1].shape)) for m in
                              self.metrics_list]
        self.metrics_funcs.sort(key=lambda metric: metric.order)

        for metric in self.metrics_funcs:
            metric.init_backend_from_model(self)
            metric.initialize()
        self.loss_and_metrics = [self.loss_func_name] + self.metrics_list
        self.loss_and_metrics_format = [self.loss_func.format] + [metric.format for metric in self.metrics_funcs]
        self.total_metrics = np.array([0] + [0 for func in self.metrics_funcs], dtype=self.dtype)
        self.tracer.define_event_types(self)
        self._initialized = True

        if self.enable_cudnn:
            assert isinstance(self.optimizer, OptimizerGPU), f"CUDA is enable but the optimizer's backend is not a GPU one ({type(self.optimizer)=})"
            self.optimizer.set_gpudirect(self.gpudirect)

        self.optimizer.initialize(self.get_all_layers(self.layers))

    def export(self):
        data = {}

        if self.model_name is not None:
            data[Parameters.MODEL_NAME] = self.model_name

        data[Parameters.LAYERS] = [
            layer.export()
            for layer in self.layers
        ]

        return data

    def import_(self, data: "dict[str, Any] | Model") -> None:
        if isinstance(data, Model):
            data = data.export()

        model_name = str(data.get(Parameters.MODEL_NAME))
        if model_name != self.model_name:
            warn(f"Importing from different models! (self: {self.model_name}, got: {model_name})", RuntimeWarning)

        for layer, data in zip(self.layers, data[Parameters.LAYERS]):
            layer.import_(data)

    def load_weights_and_bias(self, filename: str) -> None:
        """
        ARGS:
            filename: Path to the file with the weights and biases to load.
        """
        with np.load(filename, allow_pickle=True) as data:
            self.import_(data)

    def store_weights_and_bias(self, filename: str, compress=True) -> None:
        """
        ARGS:
            filename: Path to the file were the weights and biases will be stored.
        """
        save = np.savez_compressed if compress else np.savez
        save(filename, **self.export())

    def calculate_time(self) -> np.ndarray:
        # Total elapsed_time, Comp elapsed_time, Memo elapsed_time, Net elapsed_time
        total_time: np.ndarray = np.zeros((4,), dtype=np.float32)

        # Forward pass (FP)
        for layer in self.layers:
            total_time += layer.fwd_time

        if self.blocking_mpi:
            # Blocking MPI
            # Back propagation. Gradient computation (GC) and weights update (WU)
            for layer in self.layers:
                total_time += layer.bwd_time

            # Weight update (WU)
            for layer in self.layers:
                weights_size = 0 if (weights := layer.weights) is None else weights.size
                biases_size = 0 if (biases := layer.biases) is None else biases.size
                if self.comm and weights_size > 0:
                    total_time += allreduce_time(weights_size + biases_size,
                                                 self.cpu_speed, self.network_bw, self.network_lat,
                                                 self.network_alg, self.nprocs, self.dtype)
        else:
            total_time_iar: int = 0
            # Non-blocking MPI
            # Back propagation. Gradient computation (GC) and weights update (WU)
            for layer in self.layers:
                total_time += layer.bwd_time
                weights_size = 0 if (weights := layer.weights) is None else weights.size
                biases_size = 0 if (biases := layer.biases) is None else biases.size
                if self.comm and weights_size > 0:
                    time_iar = allreduce_time(weights_size + biases_size,
                                              self.cpu_speed, self.network_bw, self.network_lat,
                                              self.network_alg, self.nprocs, self.dtype)
                    total_time[3] += time_iar[3]
                    total_time_iar = max(total_time[0], total_time_iar) + time_iar[0]

            total_time[0] = max(total_time[0], total_time_iar)

        return total_time

    def _compute_metrics_funcs(self, y_pred: T, y_targ: T, loss: float, blocking=True, comm=True) -> tuple[np.ndarray, None] | tuple[None, Any]:
        loss_req: Any | None = None
        _losses: np.ndarray | None

        if y_targ.shape[0] > 0:
            metrics = [func.compute(y_pred, y_targ) for func in self.metrics_funcs]
            _losses = np.array([loss, *metrics], dtype=np.object_)
        else:
            _losses = self.total_metrics.copy()
            _losses[0] = loss

        if self.comm is not None and comm:
            assert MPI

            _losses /= self.comm_size
            if blocking:
                _losses = self.comm.allreduce(_losses, op=MPI.SUM)
            else:
                loss_req = self.comm.iallreduce(_losses, op=MPI.SUM)
        else:
            if blocking:
                pass
            else:
                raise NotImplementedError("can not compute metrics non-blocking locally")

        return _losses, loss_req

    def _update_running_average(self, curr: np.ndarray, total: np.ndarray, count: int,
                                batch_size: int, prefix="") -> tuple[np.ndarray, int, str]:
        string = ""
        total = ((curr * batch_size) + (total * count)) / (count + batch_size)
        for c in range(len(self.loss_and_metrics)):
            loss_str = self.loss_and_metrics_format[c]
            if loss_str:
                string += ("%s, " % (prefix + loss_str)) % total[c]
        string = string[:-2]
        return total, count + batch_size, string

    def _sync_x_y(self, x_batch: np.ndarray, y_batch: np.ndarray) -> tuple[T, T]:
        raise TypeError("Please, use the cpu or gpu version.")
    # --- _sync_x_y --- #

    def _sync_x_y_cpu(self, x_batch: np.ndarray, y_batch: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        self.real_batch_size = x_batch.shape[0]
        x_batch = np.asarray(x_batch, dtype=self.dtype, order='C', copy=None)
        y_batch = np.asarray(y_batch, dtype=self.dtype, order='C', copy=None)
        return x_batch, y_batch
    # --- _sync_x_y_cpu --- #

    def _sync_x_y_gpu(self, x_batch: np.ndarray, y_batch: np.ndarray) -> tuple[TensorGPU, TensorGPU]:

        # NOTE: in CUDA it's necessary to always have batches of the same size.
        local_batch_size = x_batch.shape[0]

        self.real_batch_size = local_batch_size
        if local_batch_size != 0:
            if local_batch_size != self.batch_size:
                # NOTE: if x_batch is empty (local_batch_size == 0), this will mean the end of the loop where this function is called.
                num_repetitions = ceil(self.batch_size / local_batch_size)
                x_batch = np.repeat(x_batch, num_repetitions, axis=0)[:self.batch_size]
                y_batch = np.repeat(y_batch, num_repetitions, axis=0)[:self.batch_size]
            # else: The batch has the right shape ==> Nothing to do.

            x_batch = np.asarray(x_batch, dtype=self.dtype, order='C', copy=None)
            y_batch = np.asarray(y_batch, dtype=self.dtype, order='C', copy=None)

            assert isinstance(self.layers[0].y, TensorGPU) and isinstance(self.y_batch, TensorGPU)
            self.layers[0].y.ary.set(x_batch)
            self.y_batch.ary.set(y_batch)
            x, y_targ = self.layers[0].y, self.y_batch
        else:
            x, y_targ = self.empty_x, self.empty_y_tag

        return x, y_targ
    # --- _sync_x_y_gpu --- #

    # TODO: Modify the method's name.
    def _weight_update(self, gradient=True, blocking=True):
        if blocking:
            for layer in self.layers:
                self.tracer.emit_event(PYDTNN_MDL_EVENT,
                                       layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.ALLREDUCE_DW)
                layer.reduce_weights_sync(gradient=gradient)
                self.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

        else:
            for layer in self.layers:
                self.tracer.emit_event(PYDTNN_MDL_EVENT,
                                       layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.ALLREDUCE_DW)
                layer.reduce_weights_async(gradient=gradient)
                self.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

            for layer in self.layers:
                self.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT],
                                        [layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.WAIT_DW,
                                        layer.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.OPS_ALLREDUCE_DW])
                layer.wait_allreduce_async(gradient=gradient)
                self.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT], [PYDTNN_EVENT_FINISHED, PYDTNN_EVENT_FINISHED])

    def train_dataset(self, bar_width=BAR_WIDTH) -> dict[str, list[np.ndarray]]:
        self._ensure_model_runnable()

        # If working with CUDA, self.y_batch must be in a GPU's data structure.
        if self.enable_cudnn and self.y_batch is None:
            assert gpuarray and self.cudnn_dtype
            tensor_gpu = TensorGPU(
                gpuarray.empty((self.batch_size, *self.layers[-1].shape), self.dtype),
                self.tensor_format, self.cudnn_dtype)
            self.y_batch = tensor_gpu  # type: ignore

        self.history = {lm: [] for lm in (self.loss_and_metrics + [f"val_{m}" for m in self.loss_and_metrics])}

        self.comm_nsamples = list(zip(*self.comm.allgather(self.dataset._nsamples) if self.comm else [self.dataset._nsamples]))

        terminate = False  # True: ends the following loop.
        global_terminate = False

        model_sync_count = 0
        train_batches_min = min(self.comm_nsamples[Dataset.Part.TRAIN]) / (self.batch_size * self.nprocs)
        val_batches_min = min(self.comm_nsamples[Dataset.Part.VAL]) / (self.batch_size * self.nprocs)

        for epoch in range(self.num_epochs):
            train_batch_generator, val_batch_generator = self.dataset.get_train_val_generator()
            sync_epoch = False

            train_total_loss, train_batch_count = np.zeros(len(self.loss_and_metrics)), 0
            val_total_loss, val_batch_count = np.zeros(len(self.loss_and_metrics)), 0

            if self.comm_rank == 0:
                string = ""
                fmt = "%%%dd" % (len(str(self.num_epochs)))
                epoch_string = "Epoch %s/%s" % (fmt, fmt)
                pbar = tqdm(total=self.dataset.train_nsamples, ncols=bar_width,
                            ascii=" ▁▂▃▄▅▆▇█", smoothing=0.3,
                            desc=epoch_string % (epoch + 1, self.num_epochs), unit=" samples")

            for sched in self.schedulers:
                sched.on_epoch_begin(self, self.rank)

            # --- TRAIN --- #
            for i_batch, (x_batch, y_batch, batch_size) in enumerate(train_batch_generator):
                if terminate:
                    x_batch = x_batch[:0]
                    y_batch = y_batch[:0]

                local_batch_size = x_batch.shape[0]
                sync_model = (self.model_sync_freq <= 0) or (model_sync_count % self.model_sync_freq == 0)

                if sync_model:
                    sync_epoch = True

                if model_sync_count == 0 and not self.initial_model_sync:
                    sync_model = False

                model_sync_count += 1

                if i_batch >= train_batches_min and sync_model:
                    rank_mask = self.comm.allgather(min(1, local_batch_size)) if self.comm else [min(1, local_batch_size)]
                else:
                    rank_mask = [1] * self.comm_size
                rank_avail = sum(rank_mask)

                if rank_avail <= 0 or global_terminate:
                    break

                if rank_avail < self.model_sync_min_avail:
                    sync_model = False

                self.rank_weight = self._compute_rank_weight(rank_mask, Dataset.Part.TRAIN)

                tic = timer()
                train_batch_loss = self._train_batch(x_batch, y_batch, sync_model=sync_model)
                toc = timer()

                if local_batch_size <= 0:
                    if self.comm_rank == 0:
                        pbar.set_postfix_str(s=f"{string}, waiting…", refresh=True)
                    continue

                train_total_loss, train_batch_count, string = \
                    self._update_running_average(train_batch_loss, train_total_loss, train_batch_count, batch_size)
                if self.comm_rank == 0:
                    # noinspection PyUnboundLocalVariable
                    pbar.set_postfix_str(s=string, refresh=True)
                    pbar.update(batch_size)
                    self.perf_counter.add_training_time_and_batch_size(epoch, toc - tic, batch_size)

            if self.comm_rank == 0:
                train_string = string
                for c in range(len(self.loss_and_metrics)):
                    self.history[self.loss_and_metrics[c]].append(train_total_loss[c])

            # ----------- #
            # --- VAL --- #
            # ----------- #
            for i_batch, (x_batch, y_batch, batch_size) in enumerate(val_batch_generator):
                if terminate:
                    x_batch = x_batch[:0]
                    y_batch = y_batch[:0]

                local_batch_size = x_batch.shape[0]

                sync_model = (self.model_sync_freq <= 0) or (model_sync_count % self.model_sync_freq == 0)

                if sync_model:
                    sync_epoch = True

                if model_sync_count == 0 and not self.initial_model_sync:
                    sync_model = False

                model_sync_count += 1

                if i_batch < val_batches_min:
                    rank_mask = [1] * self.comm_size
                else:
                    rank_mask = self.comm.allgather(min(1, local_batch_size)) if self.comm else [min(1, local_batch_size)]
                rank_avail = sum(rank_mask)

                if rank_avail <= 0:
                    break

                if rank_avail < self.model_sync_min_avail:
                    sync_model = False

                val_batch_loss = self._evaluate_batch(x_batch, y_batch, sync_model=False and sync_model)

                if batch_size <= 0:
                    continue

                val_total_loss, val_batch_count, string = \
                    self._update_running_average(val_batch_loss, val_total_loss, val_batch_count, batch_size, prefix="val_")
                if self.comm_rank == 0:
                    pbar.set_postfix_str(s=f"{train_string}, {string}", refresh=True)

            if self.comm_rank == 0:
                for c in range(len(self.loss_and_metrics)):
                    self.history["val_" + self.loss_and_metrics[c]].append(val_total_loss[c])

            for sched in self.schedulers:
                sched.on_epoch_end(train_total_loss, val_total_loss)
                if sched.stop_training:
                    terminate = True

            if self.comm_rank == 0:
                pbar.close()
                # Sleep for half a second to allow pbar to write its output before returning
                time.sleep(.5)

            for c in range(len(self.loss_and_metrics)):
                if not self.loss_and_metrics_format[c]:
                    print(f"{self.loss_and_metrics[c]}: {train_total_loss[c]}")
            for c in range(len(self.loss_and_metrics)):
                if not self.loss_and_metrics_format[c]:
                    print(f"val_{self.loss_and_metrics[c]}: {val_total_loss[c]}")

            if sync_epoch:
                if self.comm is not None:
                    op = MPI.LAND  # type: ignore
                    global_terminate = self.comm.allreduce(terminate, op=op)
                else:
                    global_terminate = terminate

            if global_terminate:
                break

        # Syncronize model
        if self.final_model_sync:
            self._weight_update(gradient=False, blocking=self.blocking_mpi)

        self.tracer.define_event_types(self)
        return self.history

    def _train_batch(self, x_batch: np.ndarray, y_batch: np.ndarray, sync_model=True) -> np.ndarray:
        self.mode = Model.Mode.TRAIN

        # Schedulers begin
        for sched in self.schedulers:
            sched.on_batch_begin()

        x, y_targ = self._sync_x_y(x_batch, y_batch)

        has_batch = x_batch.shape[0] > 0

        if has_batch:
            # Forward pass (FP)
            for layer in self.layers:
                self.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
                x = layer.forward(x)
                self.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            loss, dx = self.loss_func.compute(x, y_targ, self.real_batch_size)
        else:
            if y_targ.shape[0] != x_batch.shape[0]:
                raise ValueError(f"y_targ.shape[0] ({y_targ.shape[0]}) and x_batch.shape[0] ({x_batch.shape[0]}) must have the same value.")
            loss, dx = 0.0, y_targ

        total_metrics, _ = self._compute_metrics_funcs(x, y_targ, loss, comm=sync_model)
        assert total_metrics is not None
        self.total_metrics = total_metrics

        if has_batch:
            # Backward pass (BP)
            for layer in reversed(self.layers):
                self.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.BACKWARD)
                dx = layer.backward(dx)
                self.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

        if self.enable_cudnn:
            assert self.stream
            self.stream.synchronize()  # type: ignore

        # Gradient update
        if self.model_sync_freq >= 0 and sync_model:
            self._weight_update(gradient=True, blocking=self.blocking_mpi)

        if has_batch or sync_model:

            # Optimizer
            for layer in self.layers:
                self.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.UPDATE_DW)
                layer.update_weights(self.optimizer)
                self.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

        # Weight update
        if self.model_sync_freq > 0 and sync_model:
            self._weight_update(gradient=False, blocking=self.blocking_mpi)

        if self.enable_cudnn:
            for layer in self.layers:
                if layer.grad_vars:
                    layer.stream_2.synchronize()  # type: ignore

        # Schedulers end
        for sched in self.schedulers:
            sched.on_batch_end(self)

        return self.total_metrics

    def _compute_rank_weight(self, mask: list[int], part: Dataset.Part) -> float:
        match self.model_sync_participation:
            case Model.SyncParticipation.ALL:
                comm_nsamples = self.comm_nsamples[part]
            case Model.SyncParticipation.AVAIL2ALL:
                if mask[self.comm_rank]:
                    comm_nsamples = [nsamples for nsamples, mask in zip(self.comm_nsamples[part], mask) if mask]
                else:
                    return 0.0
            case _:
                raise ValueError(f"Model synchronization participation option '{self.model_sync_participation}' not recognized. Only recognized: \"{list(Model.SyncParticipation)}\"")

        min_nsamples, max_nsamples, total_nsamples = min(comm_nsamples), max(comm_nsamples), sum(comm_nsamples)
        comm_size = len(comm_nsamples)

        match self.model_sync_alg:
            case Model.SyncAlg.AVG:
                return 1.0 / comm_size
            case Model.SyncAlg.WAVG:
                return self.dataset._nsamples[part] / total_nsamples
            case Model.SyncAlg.INVAVG:
                inverse_nsamples = min_nsamples + (max_nsamples - self.dataset._nsamples[part])
                return inverse_nsamples / total_nsamples
            case _:
                raise ValueError(f"Model synchronization algorithm option '{self.model_sync_alg}' not recognized. Only recognized: \"{list(Model.SyncAlg)}\"")

    def _evaluate_batch(self, x_batch: np.ndarray, y_batch: np.ndarray, sync_model=True) -> np.ndarray:
        self.mode = Model.Mode.EVALUATE

        x, y_targ = self._sync_x_y(x_batch, y_batch)

        has_batch = x_batch.shape[0] > 0

        # Forward pass (FP)
        if has_batch:
            for i in range(len(self.layers)):
                self.tracer.emit_event(PYDTNN_MDL_EVENT, self.layers[i].id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
                x = self.layers[i].forward(x)
                self.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

            y_pred = self.layers[-1].y
            loss, _ = self.loss_func.compute(y_pred, y_targ, self.real_batch_size)
        else:
            y_pred = self.layers[-1].y
            loss = 0.0
        assert y_pred is not None

        total_metrics, _ = self._compute_metrics_funcs(y_pred, y_targ, loss, comm=sync_model)
        assert total_metrics is not None
        self.total_metrics = total_metrics

        return self.total_metrics

    def evaluate_dataset(self, bar_width=BAR_WIDTH):
        self._ensure_model_runnable()

        if self.enable_cudnn and self.y_batch is None:
            assert gpuarray and self.cudnn_dtype
            tensor_gpu = TensorGPU(
                gpuarray.empty((self.batch_size, *self.layers[-1].shape), self.dtype),
                self.tensor_format, self.cudnn_dtype)
            self.y_batch = tensor_gpu  # type: ignore

        self.comm_nsamples = list(zip(*self.comm.allgather(self.dataset._nsamples) if self.comm else [self.dataset._nsamples]))

        test_batches_min = min(self.comm_nsamples[Dataset.Part.TEST]) / (self.batch_size * self.nprocs)

        test_batch_generator = self.dataset.get_test_generator()

        if self.comm_rank == 0:
            test_total_loss, test_batch_count = np.zeros(len(self.loss_and_metrics)), 0
            pbar = tqdm(total=self.dataset.test_nsamples, ncols=bar_width,
                        ascii=" ▁▂▃▄▅▆▇█", smoothing=0.3,
                        desc="Testing", unit=" samples")

        model_sync_count = 0
        for i_batch, (x_batch, y_batch, batch_size) in enumerate(test_batch_generator):
            local_batch_size = x_batch.shape[0]

            sync_model = (self.model_sync_freq <= 0) or (model_sync_count % self.model_sync_freq == 0)

            if model_sync_count == 0 and not self.initial_model_sync:
                sync_model = False

            model_sync_count += 1

            if i_batch < test_batches_min:
                rank_mask = [1] * self.comm_size
            else:
                rank_mask = self.comm.allgather(min(1, local_batch_size)) if self.comm else [min(1, local_batch_size)]
            rank_avail = sum(rank_mask)

            if rank_avail <= 0:
                break

            if rank_avail < self.model_sync_min_avail:
                sync_model = False

            tic = timer()
            test_batch_loss = self._evaluate_batch(x_batch, y_batch, sync_model=sync_model)
            toc = timer()

            if batch_size <= 0:
                continue

            if self.comm_rank == 0:
                # noinspection PyUnboundLocalVariable
                test_total_loss, test_batch_count, string = \
                    self._update_running_average(test_batch_loss, test_total_loss, test_batch_count, batch_size, prefix="test_")
                # noinspection PyUnboundLocalVariable
                pbar.set_postfix_str(s=string, refresh=True)
                pbar.update(batch_size)
                self.perf_counter.add_testing_time_and_batch_size(self._evaluate_round, toc - tic, batch_size)

        # Increment self._evaluate_round
        self._evaluate_round += 1

        if self.comm_rank == 0:
            pbar.close()
            # Sleep for half a second to allow pbar to write its output before returning
            time.sleep(.5)



import os
import copy
import tarfile
from typing import TYPE_CHECKING

import numpy as np

from pydtnn.utils.tensor import TensorFormat
from pydtnn.datasets.dataset import Dataset
from pydtnn.utils import random


if TYPE_CHECKING:
    from pydtnn.model import Model

TRAIN_NSAMPLES = 50000
TEST_NSAMPLES = 10000
INPUT_SHAPE = (5, 40, 40)
OUTPUT_SHAPE = (2,)
IMAGES_PER_FILE = 10000


class Cyclone(Dataset):
    """
    Cyclone Dataset

    Source (SHA1): ???

    Normalize (z-score):
    offset: ???
    scale:  ???
    """

    def __init__(self, model: "Model", force_test_as_validation=False, debug=False):
        super().__init__(model, TRAIN_NSAMPLES, TEST_NSAMPLES, INPUT_SHAPE, OUTPUT_SHAPE, force_test_as_validation=force_test_as_validation, debug=debug)

    def _init_actual_data(self):
        self._src_filename = os.path.join(self.model.dataset_path, "cyclone-binary.tar.gz")
        self._xy_filenames = [
            [os.path.join("cyclone-batches-bin", f"data_batch_{x}.bin") for x in range(1, 6)],
            [],
            [os.path.join("cyclone-batches-bin", "test_batch.bin")]
        ]
        self._xy_filenames[Dataset.Part.VAL] = copy.copy(self._xy_filenames[Dataset.Part.TEST] if self.test_as_validation else self._xy_filenames[Dataset.Part.TRAIN])

        # Pregenerate GZIP indexs
        self._gzip_open(self._src_filename).close()

    def _actual_data_generator(self, part: Dataset.Part):
        xy_filenames = self._xy_filenames[part]

        if part is Dataset.Part.TRAIN and self.model.augment_shuffle:
            random.shuffle(xy_filenames)

        with self._gzip_open(self._src_filename) as g:
            with tarfile.open(fileobj=g) as t:
                for filename, offset, nsamples in self._offset2files(xy_filenames, IMAGES_PER_FILE, self._local_offset[part], self._local_nsamples[part]):
                    with t.extractfile(filename) as f:
                        x, y_classes = self._read_file(f, offset, nsamples)
                    x /= 255.0

                    y = np.zeros((*y_classes.shape, *self.output_shape), dtype=self.model.dtype, order="C")
                    self._decode_class(y, y_classes)

                    x = self.model.encode_tensor(x)

                    yield x, y

    def _read_file(self, f, offset, nsamples):
        chunk_size = np.prod(INPUT_SHAPE) + 1
        f.seek(offset * chunk_size)
        im = np.frombuffer(f.read(nsamples * chunk_size), dtype=np.uint8).reshape(nsamples, chunk_size)
        y_classes, x = im[:, 0].flatten(), im[:, 1:].reshape(nsamples, *INPUT_SHAPE).astype(self.model.dtype, order="C")
        return x, y_classes



import numpy as np

from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.multiplication import Multiplication
from pydtnn.model import Model


class MultiplicationCPU(Multiplication[np.ndarray], LayerCPU):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x1 = None
        self.x2 = None

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        # Performance model
        self.fwd_time = None
        self.bwd_time = None

    def transpose(self, x):
        return x.swapaxes(-2,-1)

    def forward(self, x1, x2):
        if self.model.mode == Model.Mode.TRAIN:
            self.x1 = x1
            self.x2 = x2
        return self.model.matmul(x1, x2)

    def backward(self, dy):
        dx1 = self.model.matmul(dy, self.transpose(self.x2))
        dx2 = self.model.matmul(self.transpose(self.x1), dy)
        return dx1, dx2



import numpy as np
from pydtnn.backends.cpu.layers.conv_2d import Conv2DCPU
from pydtnn.utils.tensor import TensorFormat, format_transpose
from pydtnn.utils.constants import Parameters

class Conv2DStandardCPU(Conv2DCPU):
    # NOTE: This is an abstract class.

    def _initializing_special_parameters(self):
        super()._initializing_special_parameters()
        match self.model.tensor_format:
                case TensorFormat.NCHW:
                    self.weights_shape = (self.co, self.ci, *self.filter_shape)
                case TensorFormat.NHWC:
                    self.weights_shape = (self.ci, *self.filter_shape, self.co)
                case _:
                    raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
    # ---

    def _export_prop(self, key: str):
        if key not in {Parameters.WEIGHTS, Parameters.DW}:
            return super()._export_prop(key)
        value = getattr(self, key)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NHWC's src: ci, kh, kw, co
                # NCHW's dst: co, ci, kh, kw
                return np.asarray(format_transpose(value, "IHWO", "OIHW"), dtype=np.float64, order="C", copy=True)
        return super()._export_prop(key)
    # -----

    def _import_prop(self, key: str, value) -> None:
        if key not in {Parameters.WEIGHTS, Parameters.DW}:
            return super()._import_prop(key, value)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NCHW's src: co, ci, kh, kw
                # NHWC's dst: ci, kh, kw, co
                ary = getattr(self, key)
                ary[:] = np.asarray(format_transpose(value, "OIHW", "IHWO"), dtype=self.model.dtype, order="C", copy=None)
                return
        return super()._import_prop(key, value)
    # -----



import gc
import time
import memray
import tempfile
from pathlib import Path
from memray._memray import compute_statistics as memray_statistics


class Profiler:
    def __init__(self):
        self.events = []

    def start(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, cls, exc, tb):
        self.stop()


class TimeProfiler(Profiler):
    def start(self):
        self._start_time = time.perf_counter()

    def stop(self):
        end_time = time.perf_counter()
        delta_time = end_time - self._start_time
        self.events.append(delta_time)


class MemoryProfiler(Profiler):
    def start(self):
        self._tmp = tempfile.mktemp()
        self._tracer = memray.Tracker(self._tmp, native_traces=True, follow_fork=True)
        gc.collect()
        self._tracer.__enter__()

    def stop(self):
        self._tracer.__exit__(None, None, None)
        stats = memray_statistics(self._tmp)
        Path(self._tmp).unlink()
        delta_memory = stats.metadata.peak_memory
        self.events.append(delta_memory)



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform
from pydtnn.layers.max_pool_2d import MaxPool2D


def vgg3do2(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    _(Conv2D(nfilters=32, filter_shape=(3, 3), padding=1, activation=Relu, weights_initializer=he_uniform))
    _(Conv2D(nfilters=32, filter_shape=(3, 3), padding=1, activation=Relu, weights_initializer=he_uniform))
    _(MaxPool2D(pool_shape=(2, 2), stride=2))
    _(Dropout(rate=0.2))
    _(Conv2D(nfilters=64, filter_shape=(3, 3), padding=1, activation=Relu, weights_initializer=he_uniform))
    _(Conv2D(nfilters=64, filter_shape=(3, 3), padding=1, activation=Relu, weights_initializer=he_uniform))
    _(MaxPool2D(pool_shape=(2, 2), stride=2))
    _(Dropout(rate=0.3))
    _(Conv2D(nfilters=128, filter_shape=(3, 3), padding=1, activation=Relu, weights_initializer=he_uniform))
    _(Conv2D(nfilters=128, filter_shape=(3, 3), padding=1, activation=Relu, weights_initializer=he_uniform))
    _(MaxPool2D(pool_shape=(2, 2), stride=2))
    _(Dropout(rate=0.4))
    _(Flatten())
    _(FC(shape=(128,), activation=Relu, weights_initializer=he_uniform))
    _(Dropout(rate=0.5))
    _(FC(shape=output_shape, activation=Softmax))

    return model



import pycuda.gpuarray as gpuarray  #type: ignore

from pydtnn.layers.input import Input
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape


class InputGPU(Input[TensorGPU], LayerGPU):

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU):
        super().initialize(prev_shape, x)
        y_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

    def forward(self, x: TensorGPU) -> TensorGPU:
        return x

    def backward(self, dy: TensorGPU) -> TensorGPU:
        return dy



import numpy as np
from pydtnn.backends.cpu.utils.relu_cython import capped_relu_cython
from pydtnn.activations.relu6 import Relu6
from pydtnn.backends.cpu.activations.activation import ActivationCPU


class Relu6CPU(Relu6[np.ndarray], ActivationCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mask: np.ndarray = None # type: ignore (will be initalized in "initialize")

    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        # NOTE: These attributes only store data, their value before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self._y = np.zeros((self.model.batch_size, *self.prev_shape), dtype=self.model.dtype, order="C")
        self._mask = np.zeros((self.model.batch_size, *self.prev_shape), dtype=np.int8, order="C")

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.y: np.ndarray = self._y[:x.shape[0], :]
        self.mask: np.ndarray = self._mask[:x.shape[0], :]
        capped_relu_cython(x.reshape(-1, copy=False, order="C"),
                           self.y.reshape(-1, copy=False, order="C"),
                           self.mask.reshape(-1, copy=False, order="C"),
                           self.cap)
        return self.y

    def backward(self, dy: np.ndarray) -> np.ndarray:
        # return dy * self.mask
        np.multiply(dy, self.mask, out=dy, dtype=self.model.dtype, order="C")
        return dy



from typing import TYPE_CHECKING
import os
import time

import numpy as np

from pydtnn.schedulers.scheduler_with_loss_or_metric import SchedulerWithLossOrMetric

if TYPE_CHECKING:
    from pydtnn.model import Model


class ModelCheckpoint(SchedulerWithLossOrMetric):
    """
    ModelCheckpoint LRScheduler
    """

    def __init__(self, loss_or_metric: str = "", epoch_save_frequency=1, verbose=True):
        super().__init__(loss_or_metric, verbose)
        self.epoch_save_frequency = epoch_save_frequency
        self.epoch_count = self.best_epoch = 0
        self.best_loss = np.inf * {True: -1, False: 1}["accuracy" in self.loss_or_metric]
        # Attributes that will be properly defined elsewhere
        self.filename: str | None = None
        self.last_filename: str | None = None

    def on_epoch_end(self, train_loss: np.ndarray[float], val_loss: np.ndarray[float]) -> None:
        idx = self._get_idx()
        self.epoch_count += 1
        loss = val_loss if self.is_val_metric else train_loss
        if ("accuracy" in self.loss_or_metric and loss[idx] > self.best_loss) or \
                ("accuracy" not in self.loss_or_metric and loss[idx] < self.best_loss):
            self.best_loss = loss[idx]
            self.best_epoch = self.epoch_count
            if self.epoch_count % self.epoch_save_frequency == 0:
                self.filename = "./model-{}-epoch-{}-{}.npz" \
                    .format(self.model.model_name, self.epoch_count, time.strftime("%Y%m%d"))
                self.model.store_weights_and_bias(self.filename)
                self.log(f"Saving model weights and bias in '{self.filename}'.")
                if self.model.comm_rank == 0 and self.last_filename is not None:
                    os.remove(self.last_filename)
                self.last_filename = self.filename

    @classmethod
    def from_model(cls, model: "Model") -> "ModelCheckpoint":
        return ModelCheckpoint(model.model_checkpoint_metric,
                               model.model_checkpoint_save_freq)



import numpy as np
import pycuda.gpuarray as gpuarray
from pycuda.compiler import SourceModule

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.metrics.kl_divergence_metric import KLDivergenceMetric
from pydtnn.backends.gpu.metrics.metric import MetricGPU


class KLDivergenceMetricGPU(KLDivergenceMetric[TensorGPU], MetricGPU):

    def __init_gpu_kernel__(self):
        module = SourceModule("""
        __global__ void kl_divergence_metric(T *y_targ, T *y_pred, T *res, int b, int n, float eps)
        {
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < b) {
                int i = 0;
                res[idx * n] = y_targ[idx * n];
                for ( i = 1; i < n; i++ ) {
                    res[idx * n + i] = fabs(y_pred[idx * n + i] * logf(fabs(y_pred[idx * n + i] / (y_targ[idx * n + i] + eps)) + eps));
                }
            }
            return;
        }
        """.replace("T", {np.float32: "float", np.float64: "double"}[self.model.dtype]))
        return module.get_function("kl_divergence_metric")

    def compute(self, y_pred, y_targ):
        threads = min(self.model.batch_size, 1024)
        blocks = max(self.model.batch_size, 1024) // threads + 1
        self.kernel(y_targ, y_pred, self.cost,
                    np.int32(self.model.batch_size), np.int32(self.shape[1]),
                    np.float32(self.eps),
                    grid=(blocks, 1, 1), block=(threads, 1, 1),
                    stream=self.model.stream)
        return gpuarray.sum(self.cost).get() / self.model.batch_size



import traceback
import types
from collections import defaultdict
from contextlib import suppress
from timeit import default_timer as timer
from typing import Hashable, Callable, Union, Any, Optional

import numpy as np
from rich import box
from rich.console import Console
from rich.table import Table
from rich.tree import Tree


class _BestOfExecution:
    """
    BestOf execution object
    """

    _names = defaultdict(lambda: 0)
    _longest_name = 0

    def __init__(self, best_of: Optional['BestOf'], execution_id: Optional[Hashable],
                 parent: Optional['_BestOfExecution']):
        self.best_of = best_of
        self.execution_id = execution_id
        self.parent = parent
        self.children: list[_BestOfExecution] = []
        self.problem_sizes = defaultdict(lambda: 0)
        if self.best_of is None:
            self.name = "Execution root"
        else:
            name = self.best_of.name
            index = _BestOfExecution._names[name]
            index += 1
            _BestOfExecution._names[name] = index
            self.name = f"{name} {index:02d}"
        _BestOfExecution._longest_name = max(_BestOfExecution._longest_name, len(self.name))
        if self.parent is not None:
            self.parent.children.append(self)
        self._blocked_by = defaultdict(lambda: defaultdict(lambda: 50))
        self._current_problem_size = None

    def __repr__(self):
        return self.name

    def block_parent(self):
        if not self.parent._is_root:
            self.parent._blocked_by[self.parent._current_problem_size][self] = 50

    def unblock_parent(self):
        if not self.parent._is_root:
            with suppress(KeyError):
                self.parent._blocked_by[self.parent._current_problem_size].pop(self)

    @property
    def is_blocked(self):
        """
        Determines whether this BestOfExecution is blocked or not. To get rid of stale blockers, each block_by counter
        is decreased by one every time this function is called.

        Returns
        -------
        It this BestOfExecution is still blocked.
        """

        # warning: as the blocked_by dictionary could be modified inside the for loop, the keys() generator is first
        #          converted to a list
        for k in list(self._blocked_by[self._current_problem_size].keys()):
            # Decrement blocking counter
            self._blocked_by[self._current_problem_size][k] -= 1
            # If the blocking counter reaches 0, release the corresponding blocker
            if self._blocked_by[self._current_problem_size][k] <= 0:
                self._blocked_by[self._current_problem_size].pop(k)
        return len(self._blocked_by[self._current_problem_size]) > 0

    def set_problem_size(self, problem_size):
        self._current_problem_size = problem_size
        self.problem_sizes[problem_size] += 1

    def print_as_table(self, time_format="6.4f"):
        self.best_of.print_as_table(execution=self, time_format=time_format)

    @property
    def summary(self):
        count = [0] * len(self.best_of.alternatives)
        # Get best_idx for this execution problem sizes
        best_idx = dict((k, self.best_of.best_idx[k])
                        for k in self.problem_sizes.keys() if k in self.best_of.best_idx)
        for idx in best_idx.values():
            count[idx] += 1
        parts = []
        total = len(best_idx)
        for i, alternative in enumerate(self.best_of.alternatives):
            if total == 0:
                parts.append(f"{alternative[0]}: ---")
            else:
                parts.append(f"{alternative[0]}: {(count[i] * 100) / total:.0f}%")
        return "\\[{}]/{}".format(" ".join(parts), total)

    @property
    def max_speedup(self):
        # Get the obtained speedups for this execution problem sizes
        all_speedups = self.best_of.speedups()
        speedups = dict((k, all_speedups[k]) for k in self.problem_sizes.keys() if k in all_speedups)
        if not len(speedups):
            return None
        total = 0
        speedup = 0
        for problem_size in speedups:  # only those with an actual speedup are considered
            count = self.problem_sizes[problem_size]
            speedup += speedups[problem_size] * count
            total += count
        return speedup / total

    @staticmethod
    def _walk_nodes(node: '_BestOfExecution', tree: Tree):
        for child in node.children:
            txt = f"{child.name:{_BestOfExecution._longest_name}s}   {child.summary}"
            max_speedup = child.max_speedup
            if max_speedup:
                txt += f" max speedup: {max_speedup:.1f}"
            branch = tree.add(txt)
            _BestOfExecution._walk_nodes(child, branch)

    def print_report(self):
        tree = Tree("BestOf execution graph")
        _BestOfExecution._walk_nodes(self, tree)
        c = Console(force_terminal=True, width=120)
        c.print(tree)

    # Protected members

    @property
    def _is_root(self):
        return self.parent is None


class BestOf:
    """
    Automatically executes one of a set of alternatives and eventually selects
    the best one, i.e., the fastest one, for each problem size.

    The alternatives are given as an array of pairs, where each pair is
    formed by:
    * the name of the alternative, and
    * the method to be called when this alternative is selected.

    All the alternative methods have to accept the same parameters and in the
    same order. For those methods with different parameters or parameters order,
    this can be enforced using an intermediate method, or a lambda function,
    that accepts the parameters in the expected order and then calls the actual
    method with its expected parameters.

    Instead of evaluating different methods, it is also possible to evaluate
    different pipelines. For doing this, the total number of stages of the
    pipeline must be specified, and the list of alternatives will provide for
    each alternative pipeline:
    * the name of the alternative pipeline, and
    * an array with the method to be called for each stage of this pipeline.

    To be able to compute the problem size of a given call, a method must be
    provided that returns the problem size as a hashable object. This
    method should accept the same parameters and in the same order as the
    methods that are going to be evaluated.
    """

    _use_first_alternative: bool = False
    _current_parents: list[_BestOfExecution] = [_BestOfExecution(best_of=None, execution_id=None, parent=None)]
    _root: _BestOfExecution = _current_parents[0]

    def __init__(self,
                 name: str,
                 alternatives: list[tuple[str, Union[Callable, list[Callable]]]],
                 get_problem_size: Callable[..., Hashable],
                 rounds: int = 10,
                 pruning_speedup: float = 10.0,
                 prune_after_round: int = 4):
        # Check parameters constraints
        assert rounds >= 1, "Rounds must be greater or equal to one."
        assert pruning_speedup > 1, "Pruning speedup must be greater than one."
        try:
            stages = len(alternatives[0][1])
        except TypeError:
            stages = 1
        if stages == 1:
            for a in alternatives:
                assert type(a[1]) in (types.FunctionType, types.LambdaType, types.BuiltinFunctionType), \
                    f"Expected a function for the '{a[0]}' alternative, got a '{type(a[1])}'."
        else:
            for a in alternatives:
                assert type(a[1]) in (list, tuple), \
                    f"Expected a list with the methods to be called for each stage of the '{a[0]}' pipeline."
                assert len(a[1]) == stages, \
                    f"Expected {stages} methods for the '{a[0]}' pipeline, received {len(a[1])}."
                for i, m in enumerate(a[1]):
                    assert type(m) in (types.FunctionType, types.LambdaType), \
                        f"Expected a function for stage {i} of the '{a[0]}' pipeline alternative."
        # Assign its initial value to each property
        self.name = name
        self.alternatives = alternatives
        self.get_problem_size = get_problem_size
        self.total_rounds = rounds
        self.stages = stages
        self.prune_after_round = prune_after_round
        self.pruning_speedup = pruning_speedup
        self.best_idx = defaultdict(lambda: -1)
        self.best_name = defaultdict(lambda: 'None')
        self.best_method: dict[Callable] = defaultdict(lambda: None)
        self.best_pipeline: dict[list[Callable]] = defaultdict(lambda: None)
        self.total_alternatives = len(self.alternatives)
        # Protected members
        self._current_round = defaultdict(lambda: 0)
        self._current_alternative = defaultdict(lambda: 0)
        self._executions: dict[list[_BestOfExecution]] = defaultdict(lambda: [])
        self._times = defaultdict(self._times_arrays)
        self._stages_times = defaultdict(self._stages_times_arrays)
        self._stages_executions = defaultdict(lambda: [0] * self.stages)
        # Set __call__() for this instance
        self._set_instance_call()

    def _times_arrays(self) -> list[list]:
        """Returns an array with n empty arrays, where n is the number of alternatives to be evaluated"""
        v = []
        for i in range(self.total_alternatives):
            v.append([])
        return v

    def _stages_times_arrays(self) -> list[list]:
        """
        Returns an array with n arrays with m Nones each, where n is the number of
        alternatives to be evaluated, and m is the number of stages.
        """
        v = []
        for i in range(self.stages):
            v.append([])
        return v

    def _register(self, execution_id) -> _BestOfExecution:
        current_parent = self._current_parents[-1]
        if execution_id in self._executions:
            for execution in self._executions[execution_id]:
                if execution.parent == current_parent:
                    return execution
        current_execution = _BestOfExecution(best_of=self, execution_id=execution_id, parent=current_parent)
        self._executions[execution_id].append(current_execution)
        return current_execution

    @classmethod
    def use_always_the_first_alternative(cls, update: bool = False):
        """
        Forces all BestOf classes to always call the first alternative,
        deactivating any competition among the different alternatives.
        """
        cls._use_first_alternative = True
        if not update:
            return
        import gc
        for obj in gc.get_objects():
            if isinstance(obj, cls):
                obj._set_instance_call()

    def _set_instance_call(self):
        if self.__class__._use_first_alternative:
            setattr(self, "__call__", self.__call_first_alternative__)
        else:
            setattr(self, "__call__", self.__call_best__)

    def __call__(self, *args, **kwargs):
        """
        __call__ is defined as an object's type. It simply calls the instance __call__.
        """
        return self.__call__(*args, **kwargs)

    def __call_first_alternative__(self, *args, **kwargs):
        """
        The first of the provided alternatives is called. The received
        parameters will be passed to it and its output will be returned.
        """

        if self.stages == 1:
            return self.alternatives[0][1](*args, **kwargs)
        else:
            # Get stage and remove stage argument from args
            args = list(args)  # Convert args to a list (so that its first element can be removed)
            stage = int(args.pop(0))
            assert stage < self.stages, \
                f"The stage number ({stage}) must be less than the specified number of stages ({self.stages})."
            return self.alternatives[0][1][stage](*args, **kwargs)

    def __call_best__(self, *args, **kwargs):
        """
        Each time this instance is called, it will call one of the different
        methods provided as alternatives. The received parameters will be passed
        to this method and its output will be returned.

        If a pipeline is being evaluated (stages > 1), the first parameter must
        provide the current stage, and the method corresponding to that stage of
        one of the given alternatives will be executed.

        Also, the execution time for a given problem size will be recorded and,
        eventually, the best method for a given problem size will be determined.

        Parameters
        ----------
        args : array
            Array of arguments to be passed to the method currently being
            evaluated. If the number of stages is greater than one, the first
            argument must be the stage that should be executed. In this case,
            this first argument will be removed from the array of arguments
            passed to the evaluated method.

        kwargs : dictionary
            Dictionary of arguments to be passed to the method currently being
            evaluated.

        Returns
        -------
        The output returned by the called method.
        """

        # Get stage
        args = list(args)  # Convert args to a list (so that its first element can be removed if in a pipeline)
        stage = int(args.pop(0)) if self.stages > 1 else 0
        assert stage < self.stages, \
            f"The stage number ({stage}) must be less than the specified number of stages ({self.stages})."
        # Get problem size and current execution
        problem_size: Any = self.get_problem_size(*args, **kwargs)
        # If best method has been already found, call it and return
        if self.stages == 1:
            with suppress(TypeError):
                return self.best_method[problem_size](*args, **kwargs)
        else:
            with suppress(TypeError):
                return self.best_pipeline[problem_size][stage](*args, **kwargs)
        # Get _current_execution_id and register this call
        current_execution_id = tuple(traceback.format_list(traceback.extract_stack()))
        current_execution = self._register(current_execution_id)
        # Set problem size and block parent until best method is found
        current_execution.set_problem_size(problem_size)
        current_execution.block_parent()
        # Set local variables for the given problem size
        current_alternative = self._current_alternative[problem_size]
        # Evaluate current alternative for current round
        BestOf._current_parents.append(current_execution)
        if self.stages == 1:
            alternative = self.alternatives[current_alternative][1]
        else:
            alternative = self.alternatives[current_alternative][1][stage]
        tic = timer()
        output = alternative(*args, **kwargs)
        elapsed_time = timer() - tic
        BestOf._current_parents.pop()
        if self.stages > 1:
            self._stages_executions[problem_size][stage] += 1
        # Stop here if any of the current execution children have not found its best alternative yet
        if current_execution.is_blocked:
            # As the blocking is asynchronous with the stage, remove any previously recorded stage times
            if self.stages > 1 and problem_size in self._stages_times:
                self._stages_times.pop(problem_size)
            # Return output
            return output
        # ---
        # If all the children have found their best alternative, record this execution and evaluate the alternatives
        # ---
        # Record execution time
        evolve = False
        round_increment = 1
        if self.stages == 1:
            self._times[problem_size][current_alternative].append(elapsed_time)
            evolve = True
        else:
            stages_times = self._stages_times[problem_size]
            stages_executions = self._stages_executions[problem_size]
            # As the unblocking is asynchronous with the stage, only record times starting from stage 0
            if stage == 0 or len(stages_times[0]) >= 1:
                stages_times[stage].append(elapsed_time)
            # Store the sum of all the stages elapsed times and change to the next alternative only if:
            #  * current stage is the last one
            #  * all the stages have been executed the same number of times,
            #  * they have been executed at least once, and
            #  * stage elapsed times have been recorded at least once.
            stages_that_met_previous_conditions = [1 for i, x in enumerate(stages_executions)
                                                   if x == stages_executions[0] >= 1 and len(stages_times[i]) >= 1]
            if stage == self.stages - 1 and len(stages_that_met_previous_conditions) == self.stages:
                medians_per_stage = [np.median(x) for x in stages_times]
                pipeline_elapsed_time = np.sum(medians_per_stage)
                self._times[problem_size][current_alternative].append(pipeline_elapsed_time)
                evolve = True
                round_increment = len(stages_times[0])
                # Remove problem_size from self._stages_executions and self._stages_times
                self._stages_executions.pop(problem_size)
                self._stages_times.pop(problem_size)
        # If evolve:
        if evolve:
            # 1) Evolve current alternative and round
            next_alternative = (current_alternative + 1) % self.total_alternatives
            current_round = self._current_round[problem_size]
            if next_alternative == 0:
                next_round = current_round + round_increment
            else:
                next_round = current_round
            # 2) If enough rounds have been performed:
            if next_round >= min(self.prune_after_round, self.total_rounds):
                best_times = [np.median(x) for x in self._times[problem_size]]
                min_time = min(best_times)
                # 2.a) Prune alternatives and ensure that the next alternative is one of remaining ones
                remaining_alternatives = [i for i, x in enumerate(best_times) if x <= min_time * self.pruning_speedup]
                if next_alternative not in remaining_alternatives:
                    for i in remaining_alternatives:
                        if i > next_alternative:
                            next_alternative = i
                            break
                    else:
                        # As no remaining alternative is greater than current next one, a round has been completed
                        next_alternative = remaining_alternatives[0]
                        next_round = current_round + round_increment
                # 2.b) Select the best method/pipeline if a new round is going to be performed and either next_round
                #      is greater than total rounds or there is only one remaining alternative
                if next_round > current_round and (next_round >= self.total_rounds or len(remaining_alternatives) == 1):
                    self.best_idx[problem_size] = best_times.index(min_time)  # first of the minimums
                    if self.stages == 1:
                        (self.best_name[problem_size],
                         self.best_method[problem_size]) = self.alternatives[self.best_idx[problem_size]]
                    else:
                        (self.best_name[problem_size],
                         self.best_pipeline[problem_size]) = self.alternatives[self.best_idx[problem_size]]
                    # Best method/pipeline set, unblock parent
                    current_execution.unblock_parent()
            # 3) Update self._current_alternative and self._current_round for the current problem size
            self._current_alternative[problem_size] = next_alternative
            self._current_round[problem_size] = next_round
        # Return output
        return output

    def best_method_has_been_found(self, *args, **kwargs):
        problem_size: Hashable = self.get_problem_size(*args, **kwargs)
        return problem_size in self.best_idx.keys()

    def medians(self):
        out = {}
        for problem_size, times in self._times.items():
            medians = []
            for i, alternative_times in enumerate(times):
                if len(alternative_times):
                    medians.append(np.median(alternative_times))
                else:
                    medians.append(np.nan)
            out[problem_size] = medians
        return out

    def speedups(self):
        out = {}
        medians = self.medians()
        for problem_size, times in self._times.items():
            best_idx = self.best_idx[problem_size]
            if best_idx != -1:
                out[problem_size] = max(medians[problem_size]) / medians[problem_size][best_idx]
        return out

    def print_as_table(self, execution=None, time_format="6.4f"):
        c = Console(force_terminal=True, width=100)
        caption = self.name if execution is None else execution.name
        t = Table(box=box.HORIZONTALS, show_header=True, header_style="blue", caption=caption)
        t.add_column("size")
        if execution is not None:
            t.add_column("count", justify="right")
        for h in [x[0] for x in self.alternatives]:
            t.add_column(str(h), justify="right")
        t.add_column("speedup", justify="right")
        medians = self.medians()
        speedups = self.speedups()
        for problem_size in self._times.keys():
            if execution is not None and problem_size not in execution.problem_sizes:
                continue
            row_contents = [""] * self.total_alternatives
            for i in range(len(self.alternatives)):
                row_contents[i] = "{0:{1}}".format(medians[problem_size][i], time_format)
            best_idx = self.best_idx[problem_size]
            if best_idx != -1:
                row_contents[best_idx] = "*[bold green]{}[/bold green]".format(row_contents[best_idx])
                row_contents.append("{:.1f}".format(speedups[problem_size]))
            else:
                row_contents.append("")
            if execution is not None:
                row_contents.insert(0, str(execution.problem_sizes[problem_size]))
            row_contents.insert(0, str(problem_size))
            t.add_row(*row_contents)
        c.print(t)

    @staticmethod
    def _walk_nodes_and_print_as_table(node: _BestOfExecution):
        for child in node.children:
            child.print_as_table()
            print()
            BestOf._walk_nodes_and_print_as_table(child)

    @staticmethod
    def print_tables():
        BestOf._walk_nodes_and_print_as_table(BestOf._root)

    @staticmethod
    def print_report():
        BestOf._root.print_report()
        print()
        BestOf.print_tables()



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def compute_dense_acc_cython(residuals: _npDT_2Dims[_np.float32],
                             dw: _npDT_2Dims[_np.float32],
                             learning_rate: float) -> _npDT_2Dims[_np.float32]:
    ...


def intersect_2d_indexes_cython(local_rows: _npDT_1Dims[_np.int32],
                                local_cols: _npDT_1Dims[_np.int32],
                                global_rows: _npDT_1Dims[_np.int32],
                                global_cols: _npDT_1Dims[_np.int32]) -> tuple[_npDT_1Dims[_np.int32], _npDT_1Dims[_np.int32]]:
    ...


def reset_residuals_cython(acc: _npDT_2Dims[_np.float32],
                           rows: _npDT_1Dims[_np.int32],
                           cols: _npDT_1Dims[_np.int32]) -> _npDT_2Dims[_np.float32]:
    ...


def update_dense_weights_cython(w: _npDT_2Dims[_np.float32],
                                u: _npDT_2Dims[_np.float32]) -> _npDT_2Dims[_np.float32]:
    ...


def update_sparsed_weights_cython(w: _npDT_2Dims[_np.float32],
                                  grads_to_update: _npDT_1Dims[_np.float32],
                                  rows_to_update: _npDT_1Dims[_np.int32],
                                  cols_to_update: _npDT_1Dims[_np.int32]) -> _npDT_2Dims[_np.float32]:
    ...


def update_sparsed_weights_mv_cython(w: _npDT_2Dims[_np.float32],
                                     grads_to_update: _npDT_1Dims[_np.float32],
                                     rows_to_update: _npDT_1Dims[_np.int32],
                                     cols_to_update: _npDT_1Dims[_np.int32],
                                     velocity: _npDT_2Dims[_np.float32],
                                     momentum: float) -> tuple[_npDT_2Dims[_np.float32], _npDT_2Dims[_np.float32]]:
    ...



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.schedulers.scheduler_with_loss_or_metric import SchedulerWithLossOrMetric

if TYPE_CHECKING:
    from pydtnn.model import Model


class ReduceLROnPlateau(SchedulerWithLossOrMetric):
    """
    ReduceLROnPlateau LRScheduler
    """

    def __init__(self, loss_or_metric: str = "", factor=0.1, patience=5, min_lr=0, verbose=True):
        # NOTE: loss_or_metric default value is "val_accuracy" in Parser.
        super().__init__(loss_or_metric, verbose)
        self.factor = factor
        self.patience = patience
        self.min_lr = min_lr
        self.best_epoch: int = 0
        self.best_loss: float = np.inf * {True: -1, False: 1}["accuracy" in self.loss_or_metric]

    def on_epoch_end(self, train_loss: np.ndarray[float], val_loss: np.ndarray[float]) -> None:
        idx = self._get_idx()
        self.epoch_count += 1
        loss = val_loss if self.is_val_metric else train_loss
        if ("accuracy" in self.loss_or_metric and loss[idx] > self.best_loss) or \
                ("accuracy" not in self.loss_or_metric and loss[idx] < self.best_loss):
            self.best_loss = loss[idx]
            self.best_epoch = self.epoch_count
        elif self.epoch_count - self.best_epoch >= self.patience \
                and self.model.optimizer.learning_rate * self.factor >= self.min_lr:
            self.model.optimizer.learning_rate *= self.factor
            self.best_epoch = self.epoch_count
            self.log("Metric '{}' did not improve for {} epochs, setting learning rate to {:.8f}."
                     .format(self.loss_or_metric, self.patience, self.model.optimizer.learning_rate))

    @classmethod
    def from_model(cls, model: "Model") -> "ReduceLROnPlateau":
        return ReduceLROnPlateau(model.reduce_lr_on_plateau_metric,
                                 model.reduce_lr_on_plateau_factor,
                                 model.reduce_lr_on_plateau_patience,
                                 model.reduce_lr_on_plateau_min_lr)



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.regression_mse import RegressionMSE


class RegressionMSECPU(RegressionMSE[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        # return np.square(y_targ - y_pred).mean()
        diff = np.subtract(y_targ, y_pred, dtype=self.model.dtype)
        np.square(diff, out=diff, dtype=self.model.dtype, casting="unsafe")
        return diff.mean(dtype=self.model.dtype)



"""OpenFHE encryption"""

import sys
import copyreg
from dataclasses import dataclass

import numpy as np

from pydtnn.libs import libcrypt

# Make sure global package is not confused with current package
_pkg = sys.path.pop(0)
try:
    import openfhe
finally:
    sys.path.insert(0, _pkg)


__all__ = (
    "Context",
)


SECURITY_LEVEL = {
    0: openfhe.SecurityLevel.HEStd_NotSet,
    128: openfhe.SecurityLevel.HEStd_128_classic,
    192: openfhe.SecurityLevel.HEStd_192_classic,
    256: openfhe.SecurityLevel.HEStd_256_classic
}


@dataclass(repr=False, eq=False, order=False, slots=True, frozen=True)
class Ciphertext[P: np.number](libcrypt.Ciphertext[openfhe.Ciphertext, P]):
    """OpenFHE ciphertext"""


class Context(libcrypt.Context[openfhe.Ciphertext]):
    """OpenFHE context"""
    _cls = Ciphertext

    def __init__(self, poly_degree: int = 13, global_scale: int = 40, security_level: int = 128) -> None:
        """Inizialize context"""
        super().__init__(poly_degree, global_scale, security_level)

        ring_dim = 2 ** self._poly_degree
        level = SECURITY_LEVEL[self._security_level]

        # Context
        parameters = openfhe.CCParamsCKKSRNS()
        parameters.SetSecurityLevel(level)
        parameters.SetRingDim(ring_dim)
        parameters.SetScalingModSize(self._global_scale)
        parameters.SetMultiplicativeDepth(0)
        self._context = openfhe.GenCryptoContext(parameters)
        self._context.Enable(openfhe.PKESchemeFeature.PKE)
        self._context.Enable(openfhe.PKESchemeFeature.KEYSWITCH)
        self._context.Enable(openfhe.PKESchemeFeature.LEVELEDSHE)

        # Keys
        keys = self._context.KeyGen()
        self._public_key = keys.publicKey
        self._private_key = keys.secretKey

    def _encrypt_chunk(self, chunk: list) -> openfhe.Ciphertext:
        """Encode list to ciphertext"""
        pack = self._context.MakeCKKSPackedPlaintext(chunk)
        cipher = self._context.Encrypt(self._public_key, pack)
        return cipher

    def _decrypt_chunk(self, chunk: openfhe.Ciphertext) -> list:
        """Decode cypertext to list"""
        pack = self._context.Decrypt(chunk, self._private_key)
        plain = pack.GetRealPackedValue()
        return plain


# Serialization
def DeserializeCryptoContext(str: bytes) -> openfhe.CryptoContext:
    """OpenFHE context deserializer"""
    return openfhe.DeserializeCryptoContextString(str, openfhe.BINARY)


def DeserializeCiphertext(str: bytes) -> openfhe.Ciphertext:
    """OpenFHE cipher text deserializer"""
    return openfhe.DeserializeCiphertextString(str, openfhe.BINARY)


def DeserializePublicKey(str: bytes) -> openfhe.PublicKey:
    """OpenFHE public key deserializer"""
    return openfhe.DeserializePublicKeyString(str, openfhe.BINARY)


def DeserializePrivateKey(str: bytes) -> openfhe.PrivateKey:
    """OpenFHE private key deserializer"""
    return openfhe.DeserializePrivateKeyString(str, openfhe.BINARY)


# Pickle support
def context_reducer(context: openfhe.CryptoContext):
    """OpenFHE context pickle reducer"""
    cls = DeserializeCryptoContext
    args = (openfhe.Serialize(context, openfhe.BINARY),)
    return (cls, args)


def ciphertext_reducer(ciphertext: openfhe.Ciphertext):
    """OpenFHE cipher text pickle reducer"""
    cls = DeserializeCiphertext
    args = (openfhe.Serialize(ciphertext, openfhe.BINARY),)
    return (cls, args)


def public_key_reducer(ciphertext: openfhe.PublicKey):
    """OpenFHE public key pickle reducer"""
    cls = DeserializePublicKey
    args = (openfhe.Serialize(ciphertext, openfhe.BINARY),)
    return (cls, args)


def private_key_reducer(ciphertext: openfhe.PrivateKey):
    """OpenFHE private key pickle reducer"""
    cls = DeserializePrivateKey
    args = (openfhe.Serialize(ciphertext, openfhe.BINARY),)
    return (cls, args)


copyreg.pickle(openfhe.CryptoContext, context_reducer)
copyreg.pickle(openfhe.Ciphertext, ciphertext_reducer)
copyreg.pickle(openfhe.PublicKey, public_key_reducer)
copyreg.pickle(openfhe.PrivateKey, private_key_reducer)



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def summ_coo_cython(self_data: _npDT_1Dims[_np.float32],
                    self_rows: _npDT_1Dims[_np.int32],
                    self_cols: _npDT_1Dims[_np.int32],
                    other_data: _npDT_1Dims[_np.float32],
                    other_rows: _npDT_1Dims[_np.int32],
                    other_cols: _npDT_1Dims[_np.int32]) -> tuple[_npDT_1Dims[_np.float32], _npDT_1Dims[_np.int32], _npDT_1Dims[_np.int32]]:
    ...


def top_threshold_selection_dense_cython(matrix: _npDT_2Dims[_np.float32],
                                         threshold: float) -> tuple[_npDT_1Dims[_np.float32], _npDT_1Dims[_np.int32], _npDT_1Dims[_np.int32]]:
    ...


def top_threshold_selection_coo_cython(values: _npDT_1Dims[_np.float32],
                                       rows: _npDT_1Dims[_np.int32],
                                       cols: _npDT_1Dims[_np.int32],
                                       threshold: float) -> tuple[_npDT_1Dims[_np.float32], _npDT_1Dims[_np.int32], _npDT_1Dims[_np.int32]]:
    ...



import numpy as np
import pycuda.gpuarray as gpuarray  # type: ignore

from pydtnn.layers.dropout import Dropout
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape
import ctypes


class DropoutGPU(Dropout[TensorGPU], LayerGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        # The following values will be initalized later:
        self.states_size: ctypes.c_size_t = None  #type: ignore
        self.space_size: ctypes.c_size_t = None  #type: ignore
        self.space: TensorGPU = None  #type: ignore
        self.states: TensorGPU = None  #type: ignore
        self.drop_desc: int | None = None
    # ----

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        # Activations y
        y_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.states_size = cudnn.cudnnDropoutGetStatesSize(self.model.cudnn_handle)
        self.space_size = cudnn.cudnnDropoutGetReserveSpaceSize(self.y.desc)

        space_gpu = gpuarray.empty((self.space_size.value,), np.byte)
        self.space = TensorGPU(space_gpu, self.model.tensor_format, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.OTHER)

        states_gpu = gpuarray.empty((self.states_size.value,), np.byte)
        self.states = TensorGPU(states_gpu, self.model.tensor_format, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.OTHER)

        self.drop_desc = cudnn.cudnnCreateDropoutDescriptor()

        cudnn.cudnnSetDropoutDescriptor(self.drop_desc, self.model.cudnn_handle, self.rate,
                                        self.states.ptr, self.states_size, seed=0)

    def forward(self, x: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        cudnn.cudnnDropoutForward(self.model.cudnn_handle, self.drop_desc,
                                  x.desc, x.ptr,
                                  self.y.desc, self.y.ptr,
                                  self.space.ptr, self.space_size.value)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        # Compute dx
        cudnn.cudnnDropoutBackward(self.model.cudnn_handle, self.drop_desc,
                                   dy.desc, dy.ptr,
                                   self.dx.desc, self.dx.ptr,
                                   self.space.ptr, self.space_size.value)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



import os
import platform

import numpy as np

from pydtnn.utils.best_of import BestOf
from pydtnn.utils import print_with_header


class BestOfProfiler:

    def __init__(self, header, best_method):
        self.header = header
        self.best_method: BestOf = best_method

    def __call__(self, *args, **kwargs):
        #
        # First run
        #
        problem_size = self.best_method.get_problem_size(*args, **kwargs)
        print(f"{problem_size}: First run (checking outputs)", sep="", end="")
        outputs = []
        for i in range(self.best_method.total_alternatives):
            outputs.append(self.best_method(*args, **kwargs))
            print(".", sep="", end="")
            if i > 0:
                if isinstance(outputs[0], np.ndarray):
                    name_0 = self.best_method.alternatives[0][0]
                    name_i = self.best_method.alternatives[i][0]
                    assert np.allclose(outputs[0], outputs[-1]), f"{name_0} and {name_i} outputs differ"
        #
        # Rest runs
        #
        print(" ", sep="", end="")
        print(f"Next runs (getting times)", sep="", end="")
        for i in range(0, (self.best_method.total_rounds - 1) * self.best_method.total_alternatives):
            if self.best_method.best_method_has_been_found(*args, **kwargs):
                break
            self.best_method(*args, **kwargs)
            print(".", sep="", end="")
        print()

    def print_results(self):
        c = Console(force_terminal=True)
        #  From IBM OpenMP documentation: If you do not set OMP_NUM_THREADS, the number of processors available is the
        #  default value to form a new team for the first encountered parallel construct.
        import multiprocessing
        num_threads = os.environ.get("OMP_NUM_THREADS", multiprocessing.cpu_count())
        msg = "{}  {}  OMP_NUM_THREADS: {}".format(self.header, platform.node(), num_threads)
        print_with_header(msg)
        self.best_method.print_as_table()



import numpy as np
import pycuda.gpuarray as gpuarray  # type: ignore
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore

from pydtnn.metrics.categorical_accuracy import CategoricalAccuracy
from pydtnn.backends.gpu.metrics.metric import MetricGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE


class CategoricalAccuracyGPU(CategoricalAccuracy[TensorGPU], MetricGPU):

    def initialize(self) -> None:
        super().initialize()
        self.cost = gpuarray.empty((self.model.batch_size,), self.model.dtype)

    def __init_gpu_kernel__(self) -> Function:
        _name = "categorical_accuracy"
        code = """
        __global__ void {name} ({T} *y_targ, {T} *y_pred, {T} *res, int b, int n)
        {{
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            const int workers = blockDim.x * gridDim.x;

            for(; idx < b; idx += workers)
            {{
                int i = 0, max = 0;
                {T} max_value = y_pred[idx * n];
                for ( i = 1; i < n; i++ ) 
                {{
                    if ( y_pred[idx * n + i] > max_value )
                    {{
                        max = i;
                        max_value = y_pred[idx * n + i];
                    }}
                }}
                res[idx] = y_targ[idx * n + max];
            }}
            return;
        }}
        """.format(T=DTYPE2CTYPE[self.model.dtype],
                   name=_name)
        
        module = SourceModule(code).get_function(_name)
        return module

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU) -> float:
        self.kernel(y_targ.ary, y_pred.ary, self.cost,
                    np.int32(self.model.batch_size), np.int32(self.shape[1]),
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)
        return gpuarray.sum(self.cost).get() * 100 / self.model.batch_size



"""Message Passing Interface"""

import os as _os

try:
    import pympi.rc as _rc
    import net_queue as _nq
except Exception:
    _rc = None
    _nq = None

if _rc:
    __all__ = _rc.__all__  # type: ignore


# Redefine protocol
proto = (
    _nq.Protocol(proto)
    if _nq and (proto := _os.environ.get("PYMPI_PROTO"))
    else None
)


def __getattr__(key):
    """Proxy all attributes to module"""
    return getattr(_rc, key)



from pathlib import Path
import warnings
import itertools
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generator, IO
from enum import IntEnum

import numpy as np
from PIL import Image
import rapidgzip

from pydtnn.utils.tensor import ChannelFormat, SampleFormat, TensorFormat, format_transpose
from pydtnn.utils import BackgroundGenerator, find_component, random
from pydtnn.utils.constants import ArrayShape

if TYPE_CHECKING:
    from pydtnn.model import Model


class Dataset(ABC):
    """
    NOTE
    - input_shape is expected to be in NCHW format
    - data_generator() is expected to be in model.dtype, normalized to [0, 1]
    - data_generator(x) is expected to be in model.tensor_format format
    - data_generator(y) is expected to be in NC format
    """

    class Part(IntEnum):
        TRAIN = 0
        VAL = 1
        TEST = 2

    def __init__(self, model: "Model", train_nsamples: int = 0, test_nsamples: int = 0, input_shape: ArrayShape = (),
                 output_shape: ArrayShape = (), force_test_as_validation=False, debug=False):

        if train_nsamples <= 0:
            raise ValueError("Dataset has no training samples!")
        elif test_nsamples <= 0:
            raise ValueError("Dataset has no test samples!")
        elif len(input_shape) <= 0:
            raise ValueError("Dataset has no input shape!")
        elif len(output_shape) <= 0:
            raise ValueError("Dataset has no output shape!")

        if len(input_shape) != 3:
            warnings.warn(f"Input shape does not have 3 dimensions ({input_shape}), it may cause issues!", RuntimeWarning)
        # if len(input_shape) == 3 and not (input_shape[0] < input_shape[2]):
        elif not (input_shape[0] < input_shape[2]):
            warnings.warn(f"Dataset input_shape {input_shape} may not be in NCHW format, regardless of model format!", RuntimeWarning)

        if len(output_shape) != 1:
            warnings.warn(f"Output shape should have 1 dimension, but it has {len(output_shape)} (Output shape: {output_shape}). This may cause issues!", RuntimeWarning)

        self.model: Model = model
        self.debug: bool = debug
        self.test_as_validation: bool = self.model.test_as_validation or force_test_as_validation
        self._nsamples: list[int] = [train_nsamples, 0, test_nsamples]

        # Compute self._nsamples[DatasetEnum.VAL]
        if self.test_as_validation:
            self._nsamples[Dataset.Part.VAL] = self._nsamples[Dataset.Part.TEST]
        else:
            self._nsamples[Dataset.Part.VAL] = min(self._nsamples[Dataset.Part.TRAIN] - self.model.nprocs,
                                                   max(self.model.nprocs,
                                                       int(self._nsamples[Dataset.Part.TRAIN] * self.model.validation_split)))
            self._nsamples[Dataset.Part.TRAIN] -= self._nsamples[Dataset.Part.VAL]

        # self.real_input_shape = tuple(input_shape)
        self.input_shape = tuple(input_shape)
        self.output_shape = tuple(output_shape)

        if self.model.transform_crop:
            crop, size = self._calculate_crop(self.input_shape[1:])  #type: ignore (The cropped input shape will be a tuple[int, int])
            self.input_shape = (self.input_shape[0], *size)

        if self.model.transform_resize:
            self.input_shape = (self.input_shape[0], self.model.transform_resize_size, self.model.transform_resize_size)

        self._initial_nsamples = [self._nsamples[Dataset.Part.TRAIN], self._nsamples[Dataset.Part.VAL], self._nsamples[Dataset.Part.TEST]]
        # Offset (in number of samples) and number of samples for the current job for each dataset part
        self._local_offset = [0] * 3
        self._local_nsamples = [0] * 3
        self._local_remaining_nsamples = [-1] * 3  # -1 is used to mark each part as not initialized

        for part in Dataset.Part.TRAIN, Dataset.Part.VAL, Dataset.Part.TEST:
            (self._local_offset[part],
             self._local_nsamples[part],
             self._nsamples[part]
             ) = self._compute_local_workload(self._nsamples[part])

        # Declare _x and _y for train, val and test dataset parts
        self._x: list[np.ndarray]
        self._y: list[np.ndarray]

        self._data_generator = self._actual_data_generator
        self._init_actual_data()

        self.x_empty_batch = np.zeros(shape=self.model.encode_shape((0, *self.input_shape)), dtype=self.model.dtype)
        self.y_empty_batch = np.zeros(shape=(0, *self.output_shape), dtype=self.model.dtype)

        if self.debug:
            self._print_report()

    def _gzip_open(self, filename: str) -> IO[bytes]:
        path = Path(filename)
        plain = path.with_suffix("")
        idx = path.with_suffix(f"{path.suffix}.idx")
        f = None

        if plain.exists():
            return open(plain, mode="rb")
        try:
            f = rapidgzip.RapidgzipFile(path, parallelization=1)
            if idx.exists():
                f.import_index(str(idx))
            else:
                f.export_index(str(idx))
        except Exception:
            if f:
                f.close()
            raise
        else:
            return f

    def export(self, split_weights: list[float] = [1]):
        """Export dataset (possibly split and rank specific)"""

        # Data generators
        gen_train = self._data_generator(Dataset.Part.TRAIN)
        gen_val = self._data_generator(Dataset.Part.VAL)
        gen_test = self._data_generator(Dataset.Part.TEST)

        # Reconstruct validation split
        if self.test_as_validation:
            gen_test = itertools.chain(gen_test, gen_val)
        else:
            gen_train = itertools.chain(gen_train, gen_val)

        # T from generators
        x_train, y_train = map(np.concat, zip(*gen_train))
        x_test, y_test = map(np.concat, zip(*gen_test))

        # Ensure dataset is in NCHW
        x_train = self.model.decode_tensor(x_train)
        x_test = self.model.decode_tensor(x_test)

        # Ensure dataset is in float64
        match self.model.dtype:
            case np.float64:
                pass
            case np.float32:
                x_train, y_train = x_train.astype(np.float64), y_train.astype(np.float64)
                x_test, y_test = x_test.astype(np.float64), y_test.astype(np.float64)
            case _:
                raise NotImplementedError(f"Unsupported model dtype {self.model.dtype}")

        # Calculate percentage splits
        total = sum(split_weights)
        split_percentage = [weight / total for weight in itertools.accumulate(split_weights)]

        # Split arrays
        np_splits = np.array(split_percentage[:-1])
        x_train = np.split(x_train, (len(x_train) * np_splits).astype(int))
        y_train = np.split(y_train, (len(y_train) * np_splits).astype(int))
        x_test = np.split(x_test, (len(x_test) * np_splits).astype(int))
        y_test = np.split(y_test, (len(y_test) * np_splits).astype(int))

        # Save arrays
        for split, (x_train, y_train, x_test, y_test) in enumerate(zip(x_train, y_train, x_test, y_test)):
            path = Path(self.model.dataset_path) / f"archive.{split}.npz"

            # Export dataset
            np.savez_compressed(path,
                                x_train=x_train,
                                y_train=y_train,
                                x_test=x_test,
                                y_test=y_test)

            # Debug information
            if self.debug:
                print(f"Export: {path}")
                print(f"x_train: {x_train.shape}")
                print(f"y_train: {y_train.shape}")
                print(f"x_test: {x_test.shape}")
                print(f"y_test: {y_test.shape}")

    @property
    def train_nsamples(self):
        return self._nsamples[Dataset.Part.TRAIN]

    @property
    def val_nsamples(self):
        return self._nsamples[Dataset.Part.VAL]

    @property
    def test_nsamples(self):
        return self._nsamples[Dataset.Part.TEST]

    def get_train_val_generator(self) -> tuple[Generator[tuple[np.ndarray, np.ndarray, int]], Generator[tuple[np.ndarray, np.ndarray, int]]]:
        return (self._batch_generator(Dataset.Part.TRAIN),
                self._batch_generator(Dataset.Part.VAL))

    def get_test_generator(self) -> Generator[tuple[np.ndarray, np.ndarray, int]]:
        return self._batch_generator(Dataset.Part.TEST)

    def _print_report(self):
        if self.model.comm_rank == 0:
            print(f"Initial nsamples:"
                  f" train: {self._initial_nsamples[Dataset.Part.TRAIN]} "
                  f" val: {self._initial_nsamples[Dataset.Part.VAL]} "
                  f" test: {self._initial_nsamples[Dataset.Part.TEST]} "
                  )
        desc = ["train", "val", "test"]
        for part in (Dataset.Part.TRAIN, Dataset.Part.VAL, Dataset.Part.TEST):
            prefix = f"{self.model.rank}: " if part is Dataset.Part.TRAIN else "   "
            print(f"{prefix}"
                  f" {desc[part]} offset: {self._local_offset[part]}"
                  f" {desc[part]} local nsamples: {self._local_nsamples[part]}"
                  f" {desc[part]} nsamples: {self._nsamples[part]}"
                  )

    def _compute_local_workload(self, nsamples: int):
        """Computes the offset (in number of samples) and the number of samples for the current rank"""

        # Reduce nsamples according to steps per epoch
        global_batch_size = self.model.batch_size * self.model.nprocs
        batches_per_worker = nsamples / global_batch_size

        if self.model.dataset_percentage != 0:
            nsamples = nsamples * self.model.dataset_percentage  # type: ignore (It's expected to receive a int as parameter and it's fine like this)

        if batches_per_worker > self.model.steps_per_epoch > 0:
            batches_per_worker = self.model.steps_per_epoch
            nsamples = batches_per_worker * global_batch_size

        # Calculate nsamples per worker
        nsamples_per_worker, big_workers = divmod(nsamples, self.model.nprocs)
        nsamples_per_big_worker = nsamples_per_worker + 1

        # Calculate local values
        if self.model.rank < big_workers:
            local_nsamples = nsamples_per_big_worker
            local_offset = self.model.rank * nsamples_per_big_worker
        else:
            local_nsamples = nsamples_per_worker
            local_offset = nsamples_per_big_worker * big_workers + nsamples_per_worker * (self.model.rank - big_workers)

        return int(local_offset), int(local_nsamples), int(nsamples)

    @abstractmethod
    def _init_actual_data(self):
        """Generates initial self._x[] and self._y[]. To be implemented in derived classes."""
        pass

    @staticmethod
    def _nchw2nhwc(x: np.ndarray) -> np.ndarray:
        return format_transpose(x, TensorFormat.NCHW, TensorFormat.NHWC)

    @staticmethod
    def _nhwc2nchw(x: np.ndarray) -> np.ndarray:
        return format_transpose(x, TensorFormat.NHWC, TensorFormat.NCHW)

    @staticmethod
    def _chw2hwc(x: np.ndarray) -> np.ndarray:
        return format_transpose(x, SampleFormat.CHW, SampleFormat.HWC)

    @staticmethod
    def _hwc2chw(x: np.ndarray) -> np.ndarray:
        return format_transpose(x, SampleFormat.HWC, SampleFormat.CHW)

    @staticmethod
    def _decode_class(y: np.ndarray, classes_list: np.ndarray) -> None:
        """Sets to 1 the corresponding entry in the 2D y array as indicated by the 1D array of classes"""
        y[np.arange(y.shape[0]), classes_list] = 1

    @staticmethod
    def _offset2files(filenames: list[str], images_per_file: int, local_offset: int, local_nsamples: int) -> list[tuple[str, int, int]]:
        i = local_offset // images_per_file
        offset_in_file = local_offset - i * images_per_file
        output = []
        while local_nsamples:
            nsamples = min(images_per_file - offset_in_file, local_nsamples)
            output.append((filenames[i], offset_in_file, nsamples))
            offset_in_file = 0
            local_nsamples -= nsamples
        return output

    def _actual_data_generator(self, part: Part) -> Generator[tuple[np.ndarray, np.ndarray]]:
        # NOTE: Yield is necessary so this function is a generator,
        #       however want to produce a empty iterable,
        #       so we return immediately without yielding
        return
        yield

    def _data_transform(self, part: Part, x: np.ndarray, y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        # NOTE: Don't modify data for the producer and ensure a mutable copy for transforms
        x, y = x.copy(), y.copy()

        if self.model.transform_crop:
            x = self._do_crop(x)

        if self.model.transform_resize:
            x = self._do_resize(x)

        if part is Dataset.Part.TRAIN:
            if self.model.augment_flip:
                x = self._do_flip_images(x)

            if self.model.augment_crop:
                x = self._do_crop_images(x)

            if self.model.augment_shuffle:
                idx = np.arange(x.shape[0])
                random.shuffle(idx)
                x, y = x[idx], y[idx]

        if self.model.normalize:
            x = self._do_normalize(x)

        return x, y

    def _actual_batch_generator(self, part: Part) -> Generator[tuple[np.ndarray, np.ndarray, int]]:
        # NOTE: global_batch_size should be MPI.reduce(x_local_batch.shape[0])
        # However to avoid communications per batch, we assume all process have our x_local_batch.shape[0]
        local_batch_size = self.model.batch_size
        global_batch_size = self.model.batch_size * self.model.nprocs

        def transform_generator():
            for x, y in self._data_generator(part):
                yield self._data_transform(part, x, y)

        generator = transform_generator()
        nsamples = self._nsamples[part]

        batch_size = 0
        batch_online = []

        while nsamples > 0:
            for x_data, y_data in generator:
                batch_online.append((x_data, y_data))
                batch_size += x_data.shape[0]
                if batch_size >= local_batch_size:
                    break

            if batch_size <= 0:
                break

            x_data, y_data = zip(*batch_online)
            x_data = np.concatenate(x_data)
            y_data = np.concatenate(y_data)
            batch_online.clear()
            batch_size = 0

            x_data, x_extra = x_data[:local_batch_size], x_data[local_batch_size:]
            y_data, y_extra = y_data[:local_batch_size], y_data[local_batch_size:]
            if extra_size := x_extra.shape[0]:
                batch_online.append((x_extra, y_extra))
                batch_size += extra_size

            while x_data.shape[0] >= local_batch_size:
                x_batch, x_data = x_data[:local_batch_size], x_data[local_batch_size:]
                y_batch, y_data = y_data[:local_batch_size], y_data[local_batch_size:]

                global_batch_size = min(nsamples, global_batch_size)
                yield x_batch[:nsamples], y_batch[:nsamples], global_batch_size
                nsamples -= global_batch_size

    def _batch_generator(self, part: Part) -> Generator[tuple[np.ndarray, np.ndarray, int]]:
        yield from BackgroundGenerator(self._actual_batch_generator(part), max_prefetch=1)
        # NOTE: The following infinite loop provides of empty batches
        #        if there are asked more batches than actually are.
        while True:
            yield self.x_empty_batch, self.y_empty_batch, 0

    def _do_normalize(self, data: np.ndarray) -> np.ndarray:
        data += self.model.normalize_offset
        data *= self.model.normalize_scale
        return data

    def _do_flip_images(self, data: np.ndarray) -> np.ndarray:
        n = data.shape[0]
        match self.model.tensor_format:
            case TensorFormat.NCHW:
                width_dim = -1
            case TensorFormat.NHWC:
                width_dim = 2
            case _:
                raise NotImplementedError(f"\"Dataset _do_flip_image\" is not implemented for \"{self.model.tensor_format}\" format.")

        limit = min(n, int(n * self.model.augment_flip_prob))
        s = np.arange(n)
        random.shuffle(s)
        s = s[:limit]
        data[s, ...] = np.flip(data[s, ...], axis=width_dim)
        return data

    def _do_crop_images(self, data: np.ndarray) -> np.ndarray:
        n, c, h, w = self.model.decode_shape(data.shape)
        crop_size = min(self.model.augment_crop_size, h, w)
        limit = min(n, int(n * self.model.augment_crop_prob))
        s = np.arange(n)
        random.shuffle(s)
        s = s[:limit]
        t = random.integers(0, h - crop_size, (limit,))
        ll = random.integers(0, w - crop_size, (limit,))
        for i, ri in enumerate(s):
            b, r = t[i] + crop_size, ll[i] + crop_size
            # batch[ri,...] = transform_resize(batch[ri,:,t[i]:b,l[i]:r], (ri.size,c,h,w))
            match self.model.tensor_format:
                case TensorFormat.NCHW:
                    data[ri, :, :t[i], :ll[i]] = 0.0
                    data[ri, :, b:, r:] = 0.0
                case TensorFormat.NHWC:
                    data[ri, :t[i], :ll[i], :] = 0.0
                    data[ri, b:, r:, :] = 0.0
                case _:
                    raise NotImplementedError(f"\"Dataset _do_crop_images\" is not implemented for \"{self.model.tensor_format}\" format.")
            data[ri, ...] = np.roll(data[ri, ...], random.integers(-t[i], (h - b)), axis=1)
            data[ri, ...] = np.roll(data[ri, ...], random.integers(-ll[i], (w - r)), axis=2)
        return data

    def _do_resize(self, data: np.ndarray) -> np.ndarray:
        data = self.model.decode_tensor(data)

        size = (self.model.transform_resize_size, self.model.transform_resize_size)
        shape = (*data.shape[:2], *size)
        N, C, H, W = shape

        new_data = np.empty(shape=shape, dtype=self.model.dtype, order="C")

        for n in range(N):
            for c in range(C):
                channel = data[n, c]
                # NOTE: PIL mode F is WH in float32
                channel = channel.transpose().astype(np.float32)
                image = Image.fromarray(channel, mode="F")
                image = image.resize(size)
                channel = np.asarray(image, dtype=np.float32)
                channel = channel.transpose().astype(self.model.dtype)
                new_data[n, c] = channel

        new_data = self.model.encode_tensor(new_data)

        return new_data
    # ---

    def _calculate_crop(self, size: tuple[int, int]) -> tuple[tuple[int, int, int, int], tuple[int, int]]:
        width, height = size
        frame_fraction = (1 - self.model.transform_crop_perc) / 2
        x_offset, y_offset = round(width * frame_fraction), round(height * frame_fraction)
        crop = (x_offset, y_offset, width - x_offset, height - y_offset)
        size = (crop[2] - crop[0], crop[3] - crop[1])
        return (crop, size)

    def _do_crop(self, data: np.ndarray) -> np.ndarray:
        data = self.model.decode_tensor(data)

        size = data.shape[2:4]
        crop, size = self._calculate_crop(size)
        shape = (*data.shape[:2], *size)
        N, C, H, W = shape

        new_data = np.empty(shape=shape, dtype=self.model.dtype, order="C")

        for n in range(N):
            for c in range(C):
                channel = data[n, c]
                # NOTE: PIL mode F is WH in float32
                channel = channel.transpose().astype(np.float32)
                image = Image.fromarray(channel, mode="F")
                image = image.crop(crop)
                channel = np.asarray(image, dtype=np.float32)
                channel = channel.transpose().astype(self.model.dtype)
                new_data[n, c] = channel

        new_data = self.model.encode_tensor(new_data)

        return new_data
    # ---
    def _load_rgb_image(self, fp: IO[bytes] | str) -> np.ndarray:
        """Transform a file-like (RGB image) to array (ndarray CHW uint8)"""
        with Image.open(fp=fp) as image:
            image = image.convert("RGB")
            array = np.asarray(image)
            # NOTE: PIL mode RGB is WHC in unit8
            array = format_transpose(array, SampleFormat.WHC, SampleFormat.CHW)
        return array

    def _load_gray_image(self, fp: IO[bytes] | str) -> np.ndarray:
        """Transform a file-like (gray-scale image) to array (ndarray CHW uint8)"""
        with Image.open(fp=fp) as image:
            image = image.convert("L")
            array = np.asarray(image)
            # NOTE: PIL mode L is WH in unit8
            array = format_transpose(array, ChannelFormat.WH, ChannelFormat.HW)
            array = array[None, ...]
        return array
    # ----


def select(name: str) -> type[Dataset]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def im2col_nchw_cython[T: _npDT](x: _npDT_4Dims[T],
                                 cols: _npDT_2Dims[T],
                                 kh: int, kw: int, ho: int, wo: int,
                                 vpadding: int, hpadding: int,
                                 vstride: int, hstride: int,
                                 vdilation: int, hdilation: int) -> None:
    """
    Args:
        x (npDT_4Dims): The 4 dimensional array (the image).
        cols (npDT_2Dims): The 2 dimensional array where the image as columns will be stored (it should be initialized as zeros).
        kh (int): Kernel's heigth.
        kw (int): Kernel's width.
        ho (int): Output's heigth.
        wo (int): Output's width.
        vpadding (int): vertical padding value.
        hpadding (int): horizontal padding value.
        vstride (int): vertical stride value.
        hstride (int): horizontal stride value.
        vdilation (int): vertical dilation value.
        hdilation (int): horizontal dilation value.
    Returns:
        Nothing. The output is stored in "cols".
    """


def col2im_nchw_cython[T: _npDT](cols: _npDT_2Dims[T],
                                 dx: _npDT_4Dims[T],
                                 n: int, c: int, h: int, w: int,
                                 kh: int, kw: int, ho: int, wo: int,
                                 vpadding: int, hpadding: int,
                                 vstride: int, hstride: int,
                                 vdilation: int, hdilation: int) -> None:
    """
    Args:
        cols (npDT_2Dims): The 2 dimensional array (the image as columns).
        dx (npDT_4Dims): The 4 dimensional array wher the image will be stored (it should be initialized as zeros).
        n (int): number of samples.
        c (int): number of channels.
        h (int): image's height.
        w (int): image's width.
        kh (int): Kernel's heigth.
        kw (int): Kernel's width.
        ho (int): Output's heigth.
        wo (int): Output's width.
        vpadding (int): vertical padding value.
        hpadding (int): horizontal padding value.
        vstride (int): vertical stride value.
        hstride (int): horizontal stride value.
        vdilation (int): vertical dilation value.
        hdilation (int): horizontal dilation value.
    Returns:
        Nothing. The output is stored in "dx".
    """
